# BioPandas
# Author: Sebastian Raschka <mail@sebastianraschka.com>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas

from biopandas.pdb import PandasPdb
import warnings
import pandas as pd
import os


TESTDATA_FILENAME = os.path.join(os.path.dirname(__file__), "data", "3eiy.pdb")
TESTDATA_FILENAME2 = os.path.join(
    os.path.dirname(__file__), "data", "4eiy_anisouchunk.pdb"
)
TESTDATA_FILENAME3 = os.path.join(
    os.path.dirname(__file__), "data", "5mtn_multichain.pdb"
)
OUTFILE = os.path.join(os.path.dirname(__file__), "data", "tmp.pdb")
OUTFILE_GZ = os.path.join(os.path.dirname(__file__), "data", "tmp.pdb.gz")

hetatm = ""
with open(TESTDATA_FILENAME, "r") as f:
    for line in f:
        if line.startswith("HETATM"):
            hetatm += line

with open(TESTDATA_FILENAME2, "r") as f:
    four_eiy = f.read()


def test_defaults():
    ppdb = PandasPdb()
    ppdb.read_pdb(TESTDATA_FILENAME)
    ppdb.to_pdb(path=OUTFILE, records=None)
    with open(TESTDATA_FILENAME, "r") as f:
        f1 = f.read()
    with open(OUTFILE, "r") as f:
        f2 = f.read()
    assert f1 == f2
    os.remove(OUTFILE)


def test_nonexpected_column():
    ppdb = PandasPdb()
    ppdb.read_pdb(TESTDATA_FILENAME)
    ppdb.df["HETATM"]["test"] = pd.Series("test", index=ppdb.df["HETATM"].index)
    with warnings.catch_warnings(record=True) as w:
        ppdb.to_pdb(path=OUTFILE, records=["HETATM"])
    with open(OUTFILE, "r") as f:
        f1 = f.read()
    os.remove(OUTFILE)
    assert f1 == hetatm


def test_records():
    """Test private _read_pdb."""
    ppdb = PandasPdb()
    ppdb.read_pdb(TESTDATA_FILENAME)
    ppdb.to_pdb(path=OUTFILE, records=["HETATM"])
    with open(OUTFILE, "r") as f:
        f1 = f.read()
    os.remove(OUTFILE)
    assert f1 == hetatm


def test_anisou():
    """Test writing ANISOU entries."""
    ppdb = PandasPdb()
    ppdb.read_pdb(TESTDATA_FILENAME2)
    ppdb.to_pdb(path=OUTFILE, records=None)
    with open(OUTFILE, "r") as f:
        f1 = f.read()
    os.remove(OUTFILE)
    assert f1 == four_eiy


def test_add_remark():
    """Test adding a REMARK entry."""
    # Add remark
    code = 3
    remark1 = 'THIS IS A HIGHLY IMPORTANT FREE-TEXT REMARK WHICH IS EXACTLY 80 CHARACTERS LONG.'
    remark2 = ''
    remark3 = 'THIS IS A NEXT MULTI-LINE INDENTED REMARK\n FOLLOWING THE BLANK REMARK.'
    ppdb = PandasPdb()
    ppdb.read_pdb(TESTDATA_FILENAME)
    n_atoms = len(ppdb.df['ATOM'])
    ppdb.add_remark(code, remark1)
    ppdb.add_remark(code, remark2)
    ppdb.add_remark(code, remark3, 5)
    ppdb.to_pdb(path=OUTFILE)

    # Test modified file contains remarks
    with open(OUTFILE, "r") as f:
        f1 = f.read()
    expected_substr = (
        "REMARK   3  OTHER REFINEMENT REMARKS: HYDROGENS HAVE BEEN ADDED IN THE RIDING   \n"
        "REMARK   3  POSITIONS                                                           \n"
        "REMARK   3 THIS IS A HIGHLY IMPORTANT FREE-TEXT REMARK WHICH IS EXACTLY 80      \n"
        "REMARK   3 CHARACTERS LONG.                                                     \n"
        "REMARK   3                                                                      \n"
        "REMARK   3      THIS IS A NEXT MULTI-LINE INDENTED REMARK                       \n"
        "REMARK   3      FOLLOWING THE BLANK REMARK.                                     \n"
        "REMARK   4                                                                      \n"
    )
    assert expected_substr in f1

    # Test number of atoms remained the same
    ppdb = PandasPdb()
    ppdb.read_pdb(OUTFILE)
    os.remove(OUTFILE)
    assert len(ppdb.df['ATOM']) == n_atoms


def test_introduce_remark():
    """Test introducing a REMARK entry to the file with no remarks."""
    # Add remark
    code = 3
    remark = 'THIS IS A HIGHLY IMPORTANT FREE-TEXT REMARK WHICH IS EXACTLY 80 CHARACTERS LONG.'
    indent = 1
    ppdb = PandasPdb()
    ppdb.read_pdb(TESTDATA_FILENAME3)
    n_atoms = len(ppdb.df['ATOM'])
    ppdb.add_remark(code, remark, indent)
    ppdb.to_pdb(path=OUTFILE)

    # Test modified file starts with new remark
    with open(OUTFILE, "r") as f:
        f1 = f.read()
    expected_prefix = (
        "REMARK   3  THIS IS A HIGHLY IMPORTANT FREE-TEXT REMARK WHICH IS EXACTLY 80     \n"
        "REMARK   3  CHARACTERS LONG.                                                    \n"
    )
    assert f1.startswith(expected_prefix)

    # Test number of atoms remained the same
    ppdb = PandasPdb()
    ppdb.read_pdb(OUTFILE)
    os.remove(OUTFILE)
    assert len(ppdb.df['ATOM']) == n_atoms

    
def test_b_factor_shift():
    """Test b_factor shifting one white space when saving the fetched pdb."""
    ppdb = PandasPdb()
    ppdb.fetch_pdb("2e28")
    ppdb.to_pdb(path=OUTFILE, records=None)
    tmp_df = ppdb.read_pdb(path=OUTFILE).df['ATOM']
    os.remove(OUTFILE)
    assert tmp_df[tmp_df["element_symbol"].isnull() | (tmp_df["element_symbol"] == '')].empty
    assert not tmp_df[tmp_df["blank_4"].isnull() | (tmp_df["blank_4"] == '')].empty




# BioPandas
# Author: Arian Jamasb <arian"jamasb.io>, Sebastian Raschka <mail@sebastianraschka.com>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas

import os

from biopandas.mmtf import PandasMmtf
from biopandas.testutils import assert_raises

TESTDATA_FILENAME = os.path.join(os.path.dirname(__file__), "data", "3eiy.mmtf")


def test_overwrite_df():
    data_path = os.path.join(os.path.dirname(__file__), "data", "3eiy.mmtf")
    pdb = PandasMmtf().read_mmtf(data_path)

    def overwrite():
        pdb.df = "bla"

    expect = (
        "Please use `PandasMmtf._df = ... ` instead\n"
        "of `PandasMmtf.df = ... ` if you are sure that\n"
        "you want to overwrite the `df` attribute."
    )

    assert_raises(AttributeError, expect, overwrite)



# BioPandas
# Author: Sebastian Raschka <mail@sebastianraschka.com>
# Author: Arian Jamasb <arian@jamasb.io>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas
import os

from biopandas.pdb import PandasPdb

TESTDATA_FILENAME = os.path.join(os.path.dirname(__file__), "data", "2jyf.pdb")


def test_get_index_df():
    df = PandasPdb().read_pdb(TESTDATA_FILENAME)
    idxs = df.get_model_start_end()["model_idx"]
    assert len(idxs) == max(idxs.astype(int))


def test_label_models():
    df = PandasPdb().read_pdb(TESTDATA_FILENAME)
    df.label_models()
    assert "model_id" in df.df["ATOM"].columns


def test_get_model():
    df = PandasPdb().read_pdb(TESTDATA_FILENAME)
    MODEL_INDEX = 1
    new_df = df.get_model(MODEL_INDEX)
    assert new_df.df["ATOM"]["model_id"].all() == MODEL_INDEX


def test_get_models():
    df = PandasPdb().read_pdb(TESTDATA_FILENAME)
    MODEL_INDICES = [1, 3, 5]

    new_df = df.get_models(MODEL_INDICES)
    assert new_df.df["ATOM"]["model_id"].all() in MODEL_INDICES



# BioPandas
# Author: Sebastian Raschka <mail@sebastianraschka.com>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas


import os
from urllib.error import HTTPError, URLError
from urllib.request import urlopen

import numpy as np
import pandas as pd
from biopandas.pdb import PandasPdb
from biopandas.testutils import assert_raises
from nose.tools import raises

TESTDATA_FILENAME = os.path.join(os.path.dirname(__file__), "data", "3eiy.pdb")
TESTDATA_FILENAME2 = os.path.join(
    os.path.dirname(__file__), "data", "4eiy_anisouchunk.pdb"
)
TESTDATA_FILENAME_GZ = os.path.join(os.path.dirname(__file__), "data", "3eiy.pdb.gz")
TESTDATA_FILENAME_AF2_V4 = os.path.join(
    os.path.dirname(__file__), "data", "AF-Q5VSL9-F1-model_v4.pdb"
)

TESTDATA_FILENAME_AF2_V3 = os.path.join(
    os.path.dirname(__file__), "data", "AF-Q5VSL9-F1-model_v3.pdb"
)

ATOM_DF_COLUMNS = [
    "record_name",
    "atom_number",
    "blank_1",
    "atom_name",
    "alt_loc",
    "residue_name",
    "blank_2",
    "chain_id",
    "residue_number",
    "insertion",
    "blank_3",
    "x_coord",
    "y_coord",
    "z_coord",
    "occupancy",
    "b_factor",
    "blank_4",
    "segment_id",
    "element_symbol",
    "charge",
    "line_idx",
]

ANISOU_DF_COLUMNS = [
    "record_name",
    "atom_number",
    "blank_1",
    "atom_name",
    "alt_loc",
    "residue_name",
    "blank_2",
    "chain_id",
    "residue_number",
    "insertion",
    "blank_3",
    "U(1,1)",
    "U(2,2)",
    "U(3,3)",
    "U(1,2)",
    "U(1,3)",
    "U(2,3)",
    "blank_4",
    "element_symbol",
    "charge",
    "line_idx",
]

with open(TESTDATA_FILENAME, "r") as f:
    three_eiy = f.read()

with open(TESTDATA_FILENAME2, "r") as f:
    four_eiy = f.read()

with open(TESTDATA_FILENAME_AF2_V4, "r") as f:
    af_test_struct_v4 = f.read()

with open(TESTDATA_FILENAME_AF2_V3, "r") as f:
    af_test_struct_v3 = f.read()


def test__read_pdb():
    """Test private _read_pdb"""
    ppdb = PandasPdb()
    path, txt = ppdb._read_pdb(TESTDATA_FILENAME)
    print(txt)
    assert txt == three_eiy


def test__read_pdb_raises():
    """Test private _read_pdb:
    Test if ValueError is raised for wrong file formats."""

    expect = (
        "Wrong file format; allowed file formats are " ".pdb, .pdb.gz, .ent, .ent.gz"
    )

    def run_code_1():
        PandasPdb()._read_pdb("protein.mol2")

    assert_raises(ValueError, expect, run_code_1)

    def run_code_2():
        PandasPdb()._read_pdb("protein.mol2.gz")

    assert_raises(ValueError, expect, run_code_2)


def test_fetch_pdb():
    """Test fetch_pdb"""

    try:
        ppdb = PandasPdb()
        url, txt = ppdb._fetch_pdb("3eiy")
    except HTTPError:
        url, txt = None, None
    except ConnectionResetError:
        url, txt = None, None

    if txt:  # skip if PDB down
        txt[:100] == three_eiy[:100]
        ppdb.fetch_pdb("3eiy")
        assert ppdb.pdb_text == txt
        assert ppdb.pdb_path == "https://files.rcsb.org/download/3eiy.pdb"


def test_fetch_af2():
    """Test fetch_pdb"""
    # Check latest release
    try:
        ppdb = PandasPdb()
        url, txt = ppdb._fetch_af2("Q5VSL9", af2_version=4)
    except HTTPError:
        url, txt = None, None
    except ConnectionResetError:
        url, txt = None, None

    if txt:  # skip if AF2 DB down
        txt[:100] == af_test_struct_v4[:100]
        ppdb.fetch_pdb(uniprot_id="Q5VSL9", source="alphafold2-v4")
        assert ppdb.pdb_text == txt
        assert (
            ppdb.pdb_path
            == "https://alphafold.ebi.ac.uk/files/AF-Q5VSL9-F1-model_v4.pdb"
        )

    # Check legacy release
    try:
        ppdb = PandasPdb()
        url, txt = ppdb._fetch_af2("Q5VSL9", af2_version=3)
    except HTTPError:
        url, txt = None, None
    except ConnectionResetError:
        url, txt = None, None

    if txt:  # skip if AF2 DB down
        txt[:100] == af_test_struct_v3[:100]
        ppdb.fetch_pdb(uniprot_id="Q5VSL9", source="alphafold2-v3")
        assert ppdb.pdb_text == txt
        assert (
            ppdb.pdb_path
            == "https://alphafold.ebi.ac.uk/files/AF-Q5VSL9-F1-model_v3.pdb"
        )


def test__read_pdb_gz():
    """Test public _read_pdb with gzip files"""
    ppdb = PandasPdb()
    path, txt = ppdb._read_pdb(TESTDATA_FILENAME_GZ)
    assert txt == three_eiy


def test__construct_df():
    """Test pandas dataframe construction"""
    ppdb = PandasPdb()
    dfs = ppdb._construct_df(three_eiy.splitlines())
    assert set(dfs.keys()) == {"OTHERS", "ATOM", "ANISOU", "HETATM"}
    assert set(dfs["ATOM"].columns) == set(ATOM_DF_COLUMNS)
    assert set(dfs["HETATM"].columns) == set(ATOM_DF_COLUMNS)
    assert set(dfs["ANISOU"].columns) == set(ANISOU_DF_COLUMNS)
    exp = pd.Series(
        np.array(
            [
                "ATOM",
                1,
                "",
                "N",
                "",
                "SER",
                "",
                "A",
                2,
                "",
                "",
                2.527,
                54.656,
                -1.667,
                1.0,
                52.73,
                "",
                "",
                "N",
                None,
                609,
            ]
        ),
        index=[
            "record_name",
            "atom_number",
            "blank_1",
            "atom_name",
            "alt_loc",
            "residue_name",
            "blank_2",
            "chain_id",
            "residue_number",
            "insertion",
            "blank_3",
            "x_coord",
            "y_coord",
            "z_coord",
            "occupancy",
            "b_factor",
            "blank_4",
            "segment_id",
            "element_symbol",
            "charge",
            "line_idx",
        ],
    )
    assert exp.equals(dfs["ATOM"].loc[0, :])


def test_read_pdb():
    """Test public read_pdb"""
    ppdb = PandasPdb()
    ppdb.read_pdb(TESTDATA_FILENAME)
    assert ppdb.pdb_text == three_eiy
    assert ppdb.code == "3eiy", ppdb.code
    assert ppdb.pdb_path == TESTDATA_FILENAME


def test_read_pdb_from_list():
    """Test public read_pdb_from_list"""

    for pdb_text, code in zip([three_eiy, four_eiy], ["3eiy", "4eiy"]):
        ppdb = PandasPdb()
        ppdb.read_pdb_from_list(pdb_text.splitlines(True))
        assert ppdb.pdb_text == pdb_text
        assert ppdb.code == code
        assert ppdb.pdb_path == ""


def test_anisou_input_handling():
    """Test public read_pdb"""
    ppdb = PandasPdb()
    ppdb.read_pdb(TESTDATA_FILENAME2)
    assert ppdb.pdb_text == four_eiy
    assert ppdb.code == "4eiy", ppdb.code


@raises(AttributeError)
def test_get_exceptions():
    ppdb = PandasPdb()
    ppdb.read_pdb(TESTDATA_FILENAME)
    ppdb.get("main-chai")


def test_get_all():
    ppdb = PandasPdb()
    ppdb.read_pdb(TESTDATA_FILENAME)
    for i in ["c-alpha", "hydrogen", "main chain"]:
        ppdb.get(i)


def test_get_df():
    ppdb = PandasPdb()
    ppdb.read_pdb(TESTDATA_FILENAME)

    shape = ppdb.get("c-alpha").shape
    assert shape == (174, 21), shape

    shape = ppdb.get("hydrogen", invert=True, records=("ATOM",)).shape
    assert shape == (1330, 21), shape

    # deprecated use of string
    shape = ppdb.get("hydrogen", invert=True, records="ATOM").shape
    assert shape == (1330, 21), shape

    shape = ppdb.get("hydrogen").shape
    assert shape == (0, 21), shape

    shape = ppdb.get("main chain", records=("ATOM",)).shape
    assert shape == (696, 21), shape

    shape = ppdb.get("heavy", records=("ATOM",)).shape
    assert shape == (1330, 21), shape

    shape = ppdb.get("carbon", records=("ATOM",)).shape
    assert shape == (857, 21), shape



""" Class for working with PDB files"""

# BioPandas
# Author: Sebastian Raschka <mail@sebastianraschka.com>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas
from __future__ import annotations

import gzip
import sys
import warnings
import textwrap
from copy import deepcopy
from io import StringIO
from typing import List, Optional
from urllib.error import HTTPError, URLError
from urllib.request import urlopen
from warnings import warn

import numpy as np
import pandas as pd
from looseversion import LooseVersion

from .engines import amino3to1dict, pdb_df_columns, pdb_records
from biopandas.constants import ATOMIC_MASSES

pd_version = LooseVersion(pd.__version__)


class PandasPdb(object):
    """
    Object for working with Protein Databank structure files.

    Attributes
    ----------
    df : dict
        Dictionary storing pandas DataFrames for PDB record sections.
        The dictionary keys are {'ATOM', 'HETATM', 'ANISOU', 'OTHERS'}
        where 'OTHERS' contains all entries that are not parsed as
        'ATOM', 'HETATM', or 'ANISOU'.

    pdb_text : str
        PDB file contents in raw text format.

    pdb_path : Union[str, os.PathLike]
        Location of the PDB file that was read in via `read_pdb`
        or URL of the page where the PDB content was fetched from
        if `fetch_pdb` was called.

    header : str
        PDB file description.

    code : str
        PDB code

    """

    def __init__(self):
        self._df = {}
        self.pdb_text = ""
        self.header = ""
        self.code = ""
        self._get_dict = {}
        self.pdb_path = ""

    @property
    def df(self):
        """Access dictionary of pandas DataFrames for PDB record sections."""
        return self._df

    @df.setter
    def df(self, value):
        """Assign a new value to the pandas DataFrame"""
        raise AttributeError(
            "Please use `PandasPdb._df = ... ` instead\n"
            "of `PandasPdb.df = ... ` if you are sure that\n"
            "you want to overwrite the `df` attribute."
        )
        # self._df = value

    def read_pdb(self, path):
        """Read PDB files (unzipped or gzipped) from local drive

        Attributes
        ----------
        path : str
            Path to the PDB file in .pdb format or gzipped format (.pdb.gz).

        Returns
        ---------
        self

        """
        self.pdb_path, self.pdb_text = self._read_pdb(path=str(path))
        self._df = self._construct_df(pdb_lines=self.pdb_text.splitlines(True))
        self.header, self.code = self._parse_header_code()
        return self

    def read_pdb_from_list(self, pdb_lines):
        """Reads PDB file from a list into DataFrames

        Attributes
        ----------
        pdb_lines : list
            A list of lines containing the pdb file contents.

        Returns
        ---------
        self

        """
        self.pdb_text = "".join(pdb_lines)
        self._df = self._construct_df(pdb_lines)
        self.header, self.code = self._parse_header_code()
        return self

    def fetch_pdb(self, pdb_code: Optional[str] = None, uniprot_id: Optional[str] = None, source: str = "pdb"):
        """Fetches PDB file contents from the Protein Databank at rcsb.org or AlphaFold database
        at https://alphafold.ebi.ac.uk/.
.

        Parameters
        ----------
        pdb_code : str, optional
            A 4-letter PDB code, e.g., `"3eiy"` to retrieve structures from the PDB.
            Defaults to `None`.

        uniprot_id : str, optional
            A UniProt Identifier, e.g., `"Q5VSL9"` to retrieve structures from the AF2 database.
            Defaults to `None`.

        source : str
            The source to retrieve the structure from 
            (`"pdb"`, `"alphafold2-v3"`, `"alphafold2-v4"`(latest)).
            Defaults to `"pdb"`.

        Returns
        ---------
        self

        """
        # Sanitize input
        invalid_input_identifier_1 = pdb_code is None and uniprot_id is None
        invalid_input_identifier_2 = pdb_code is not None and uniprot_id is not None
        invalid_input_combination_1 = uniprot_id is not None and source == "pdb"
        invalid_input_combination_2 = pdb_code is not None and source in {
            "alphafold2-v3", "alphafold2-v4"}

        if invalid_input_identifier_1 or invalid_input_identifier_2:
            raise ValueError("Please provide either a PDB code or a UniProt ID.")

        if invalid_input_combination_1:
            raise ValueError("Please use a 'pdb_code' instead of 'uniprot_id' for source='pdb'.")
        elif invalid_input_combination_2:
            raise ValueError(f"Please use a 'uniprot_id' instead of 'pdb_code' for source={source}.")

        if source == "alphafold2-v3":
            af2_version = 3
            self.pdb_path, self.pdb_text = self._fetch_af2(uniprot_id, af2_version)
        elif source == "alphafold2-v4":
            af2_version = 4
            self.pdb_path, self.pdb_text = self._fetch_af2(uniprot_id, af2_version)
        elif source == "pdb":
            self.pdb_path, self.pdb_text = self._fetch_pdb(pdb_code)
        else:
            raise ValueError(f"Invalid source: {source}."
                " Please use one of 'pdb' or 'alphafold2-v3' or 'alphafold2-v4'.")

        self._df = self._construct_df(pdb_lines=self.pdb_text.splitlines(True))
        return self

    def get(self, s, df=None, invert=False, records=("ATOM", "HETATM")):
        """Filter PDB DataFrames by properties

        Parameters
        ----------
        s : str  in {'main chain', 'hydrogen', 'c-alpha', 'heavy'}
            String to specify which entries to return.

        df : pandas.DataFrame, default: None
            Optional DataFrame to perform the filter operation on.
            If df=None, filters on self.df['ATOM'].

        invert : bool, default: True
            Inverts the search query. For example if s='hydrogen' and
            invert=True, all but hydrogen entries are returned.

        records : iterable, default: ('ATOM', 'HETATM')
            Specify which record sections to consider. For example, to consider
            both protein and ligand atoms, set `records=('ATOM', 'HETATM')`.
            This setting is ignored if `df` is not set to None.
            For downward compatibility, a string argument is still supported
            but deprecated and will be removed in future versions.

        Returns
        --------
        df : pandas.DataFrame
            Returns a DataFrame view on the filtered entries.

        """
        if isinstance(records, str):
            warnings.warn(
                "Using a string as `records` argument is "
                "deprecated and will not be supported in future"
                " versions. Please use a tuple or"
                " other iterable instead",
                DeprecationWarning,
            )
            records = (records,)

        if not self._get_dict:
            self._get_dict = self._init_get_dict()
        if s not in self._get_dict.keys():
            raise AttributeError(f"s must be in {self._get_dict.keys()}")
        if not df:
            df = pd.concat(objs=[self.df[i] for i in records])
        return self._get_dict[s](df, invert=invert)

    def impute_element(self, records=("ATOM", "HETATM"), inplace=False):
        """Impute element_symbol from atom_name section.

        Parameters
        ----------
        records : iterable, default: ('ATOM', 'HETATM')
            Coordinate sections for which the element symbols should be
            imputed.

        inplace : bool, default: False
            Performs the operation in-place if True and returns a copy of the
            PDB DataFrame otherwise.

        Returns
        ---------
        DataFrame

        """
        if inplace:
            t = self.df
        else:
            t = self.df.copy()
            for d in self.df:
                t[d] = self.df[d].copy()

        for sec in records:
            t[sec]["element_symbol"] = t[sec][["atom_name", "element_symbol"]].apply(
                lambda x: x[0][1] if len(x[1]) == 3 else x[0][0], axis=1
            )
        return t
    
    def add_remark(self, code, text='', indent=0):
        """Add custom REMARK entry.

        The remark will be inserted to preserve the ordering of REMARK codes, i.e. if the code is
        `n` it will be added after all remarks with codes less or equal to `n`. If the object does
        not store any remarks the remark will be inserted right before the first of ATOM, HETATM or 
        ANISOU records.

        Parameters
        ----------
        code : int
            REMARK code according to PDB standards.

        text : str
            The text of the remark. If the text does not fit into a single line it will be wrapped
            into multiple lines of REMARK entries. Likewise, if the text contains new line 
            characters it will be split accordingly.
        
        indent : int, default: 0
            Number of white spaces inserted before the text of the remark.

        Returns
        ---------
        Nothing

        """
        # Prepare info from self
        if 'OTHERS' in self.df:
            df_others = self.df['OTHERS']
        else:
            df_others = pd.DataFrame(columns=['record_name', 'entry', 'line_idx'])
        record_types = list(filter(lambda x: x in self.df, ['ATOM', 'HETATM', 'ANISOU']))
        remarks = df_others[df_others['record_name'] == 'REMARK']['entry']

        # Find index and line_idx where to insert the remark to preserve remark code order
        if len(remarks):
            remark_codes = remarks.apply(lambda x: x.split(maxsplit=1)[0]).astype(int)
            insertion_pos = remark_codes.searchsorted(code, side='right')
            if insertion_pos < len(remark_codes):  # Remark in the middle
                insertion_idx = remark_codes.index[insertion_pos]
                insertion_line_idx = df_others.loc[insertion_idx]['line_idx']
            else:  # Last remark
                insertion_idx = len(remark_codes)
                insertion_line_idx = df_others['line_idx'].iloc[-1] + 1
        else:  # First remark
            insertion_idx = 0
            insertion_line_idx = min([self.df[r]['line_idx'].min() for r in record_types])

        # Wrap remark to fit into 80 characters per line and add indentation
        wrapper = textwrap.TextWrapper(width=80 - (11 + indent))
        lines = sum([wrapper.wrap(l.strip()) or [' '] for l in text.split('\n')], [])
        lines = list(map(lambda x: f'{code:4} ' +  indent*' ' + x, lines))

        # Shift data frame indices and row indices to create space for the remark
        # Create space in OTHERS
        line_idx = df_others['line_idx'].copy()
        line_idx[line_idx >= insertion_line_idx] += len(lines)
        df_others['line_idx'] = line_idx
        index = pd.Series(df_others.index.copy())
        index[index >= insertion_idx] += len(lines)
        df_others.index = index
        # Shift all other record types that follow inserted remark
        for records in record_types:
            df_records = self.df[records]
            if not insertion_line_idx > df_records['line_idx'].max():
                df_records['line_idx'] += len(lines)

        # Put remark into 'OTHERS' data frame
        df_remark = {
            idx: ['REMARK', line, line_idx]
            for idx, line, line_idx in zip(
                range(insertion_idx, insertion_idx + len(lines)),
                lines,
                range(insertion_line_idx, insertion_line_idx + len(lines)),
            )
        }
        df_remark = pd.DataFrame.from_dict(df_remark, orient='index', columns=df_others.columns)
        self.df['OTHERS'] = pd.concat([df_others, df_remark]).sort_index()

    @staticmethod
    def rmsd(df1, df2, s=None, invert=False, decimals=4):
        """Compute the Root Mean Square Deviation between molecules.

        Parameters
        ----------
        df1 : pandas.DataFrame
            DataFrame with HETATM, ATOM, and/or ANISOU entries.

        df2 : pandas.DataFrame
            Second DataFrame for RMSD computation against df1. Must have the
            same number of entries as df1.

        s : {'main chain', 'hydrogen', 'c-alpha', 'heavy', 'carbon'} or None,
            default: None
            String to specify which entries to consider. If None, considers
            all atoms for comparison.

        invert : bool, default: False
            Inverts the string query if true. For example, the setting
            `s='hydrogen', invert=True` computes the RMSD based on all
            but hydrogen atoms.

        decimals : int, default: 4
            Specifies the number of decimal places to round the final value to.

        Returns
        ---------
        rmsd : float
            Root Mean Square Deviation between df1 and df2

        """
        if df1.shape[0] != df2.shape[0]:
            raise AttributeError("DataFrames have unequal lengths")
        get_dict = PandasPdb._init_get_dict()
        if s:
            if s not in get_dict.keys():
                raise AttributeError(f"s must be in {get_dict.keys()} or None")
            df1 = get_dict[s](df1, invert=invert)
            df2 = get_dict[s](df2, invert=invert)

        total = (
            (df1["x_coord"].values - df2["x_coord"].values) ** 2
            + (df1["y_coord"].values - df2["y_coord"].values) ** 2
            + (df1["z_coord"].values - df2["z_coord"].values) ** 2
        )
        return round((total.sum() / df1.shape[0]) ** 0.5, decimals)

    @staticmethod
    def _init_get_dict():
        """Initialize dictionary for filter operations."""
        return {
            "main chain": PandasPdb._get_mainchain,
            "hydrogen": PandasPdb._get_hydrogen,
            "c-alpha": PandasPdb._get_calpha,
            "carbon": PandasPdb._get_carbon,
            "heavy": PandasPdb._get_heavy,
        }

    @staticmethod
    def _read_pdb(path):
        """Read PDB file from local drive."""
        r_mode = "r"
        if path.endswith((".pdb", ".ent")):
            openf = open
        elif path.endswith(("pdb.gz", ".ent.gz")):
            r_mode = "rb"
            openf = gzip.open
        else:
            allowed_formats = ", ".join((".pdb", ".pdb.gz", ".ent", ".ent.gz"))
            raise ValueError(
                f"Wrong file format; allowed file formats are {allowed_formats}"
            )

        with openf(path, r_mode) as f:
            txt = f.read()

        if path.endswith(".gz"):
            txt = (
                txt.decode("utf-8") if sys.version_info[0] >= 3 else txt.encode("ascii")
            )
        return path, txt

    @staticmethod
    def _fetch_pdb(pdb_code):
        """Load PDB file from rcsb.org."""
        txt = None
        url = f"https://files.rcsb.org/download/{pdb_code.lower()}.pdb"
        try:
            response = urlopen(url)
            txt = response.read()
            txt = (
                txt.decode("utf-8") if sys.version_info[0] >= 3 else txt.encode("ascii")
            )
        except HTTPError as e:
            print(f"HTTP Error {e.code}")
        except URLError as e:
            print(f"URL Error {e.args}")
        return url, txt

    @staticmethod
    def _fetch_af2(uniprot_id: str, af2_version: int = 3):
        """Load PDB file from https://alphafold.ebi.ac.uk/."""
        txt = None
        url = f"https://alphafold.ebi.ac.uk/files/AF-{uniprot_id.upper()}-F1-model_v{af2_version}.pdb"
        try:
            response = urlopen(url)
            txt = response.read()
            txt = txt.decode('utf-8') if sys.version_info[0] >= 3 else txt.encode('ascii')
        except HTTPError as e:
            print(f'HTTP Error {e.code}')
        except URLError as e:
            print(f'URL Error {e.args}')
        return url, txt

    def _parse_header_code(self):
        """Extract header information and PDB code."""
        code, header = "", ""
        if "OTHERS" in self.df:

            header = self.df["OTHERS"][self.df["OTHERS"]["record_name"] == "HEADER"]
            if not header.empty:
                header = header["entry"].values[0]
                s = header.split()
                if s:
                    code = s[-1].lower()
        return header, code

    @staticmethod
    def _get_mainchain(df, invert):
        """Return only main chain atom entries from a DataFrame"""
        if invert:
            mc = df[
                (df["atom_name"] != "C")
                & (df["atom_name"] != "O")
                & (df["atom_name"] != "N")
                & (df["atom_name"] != "CA")
            ]
        else:
            mc = df[
                (df["atom_name"] == "C")
                | (df["atom_name"] == "O")
                | (df["atom_name"] == "N")
                | (df["atom_name"] == "CA")
            ]
        return mc

    @staticmethod
    def _get_hydrogen(df, invert):
        """Return only hydrogen atom entries from a DataFrame"""
        if invert:
            return df[(df["element_symbol"] != "H")]
        else:
            return df[(df["element_symbol"] == "H")]

    @staticmethod
    def _get_heavy(df, invert):
        """Return only heavy atom entries from a DataFrame"""
        if invert:
            return df[df["element_symbol"] == "H"]
        else:
            return df[df["element_symbol"] != "H"]

    @staticmethod
    def _get_calpha(df, invert):
        """Return c-alpha atom entries from a DataFrame"""
        if invert:
            return df[df["atom_name"] != "CA"]
        else:
            return df[df["atom_name"] == "CA"]

    @staticmethod
    def _get_carbon(df, invert):
        """Return carbon atom entries from a DataFrame"""
        if invert:
            return df[df["element_symbol"] != "C"]
        else:
            return df[df["element_symbol"] == "C"]

    @staticmethod
    def _construct_df(pdb_lines):
        """Construct DataFrames from list of PDB lines."""
        valids = tuple(pdb_records.keys())
        line_lists = {r: [] for r in valids}
        line_lists["OTHERS"] = []
        for line_num, line in enumerate(pdb_lines):
            if line.strip():
                if line.startswith(valids):
                    record = line[:6].rstrip()
                    line_ele = ["" for _ in range(len(pdb_records[record]) + 1)]
                    for idx, ele in enumerate(pdb_records[record]):
                        line_ele[idx] = line[ele["line"][0]: ele["line"][1]].strip()
                    line_ele[-1] = line_num
                    line_lists[record].append(line_ele)
                else:
                    line_lists["OTHERS"].append(
                        [line[:6].rstrip(), line[6:-1].rstrip(), line_num]
                    )

        dfs = {}
        for r in line_lists.items():
            df = pd.DataFrame(
                r[1], columns=[c["id"] for c in pdb_records[r[0]]] + ["line_idx"]
            )
            for c in pdb_records[r[0]]:
                try:
                    df[c["id"]] = df[c["id"]].astype(c["type"])
                except ValueError:
                    # expect ValueError if float/int columns are empty strings
                    df[c["id"]] = pd.Series(np.nan, index=df.index)

            dfs[r[0]] = df

        # issue a warning if no atoms have been loaded
        if (len(dfs["ATOM"]) + len(dfs["HETATM"])) == 0:
            warnings.warn(
                "No ATOM/HETATM entries have been loaded. "
                "Is the input file/text in the pdb format?"
            )

        return dfs

    def amino3to1(self, record="ATOM", residue_col="residue_name", fillna="?"):
        """Creates 1-letter amino acid codes from DataFrame

        Non-canonical amino-acids are converted as follows:
        ASH (protonated ASP) => D
        CYX (disulfide-bonded CYS) => C
        GLH (protonated GLU) => E
        HID/HIE/HIP (different protonation states of HIS) = H
        HYP (hydroxyproline) => P
        MSE (selenomethionine) => M

        Parameters
        ----------
        record : str, default: 'ATOM'
            Specifies the record DataFrame.
        residue_col : str,  default: 'residue_name'
            Column in `record` DataFrame to look for 3-letter amino acid
            codes for the conversion.
        fillna : str, default: '?'
            Placeholder string to use for unknown amino acids.

        Returns
        ---------
        pandas.DataFrame : Pandas DataFrame object consisting of two columns,
            `'chain_id'` and `'residue_name'`, where the former contains
            the chain ID of the amino acid and the latter
            contains the 1-letter amino acid code, respectively.

        """
        tmp = self.df[record]
        cmp = "placeholder"
        indices = []

        residue_number_insertion = tmp["residue_number"].astype(str) + tmp["insertion"]

        for num, ind in zip(residue_number_insertion, np.arange(tmp.shape[0])):
            if num != cmp:
                indices.append(ind)
            cmp = num

        transl = tmp.iloc[indices][residue_col].map(amino3to1dict).fillna(fillna)

        return pd.concat((tmp.iloc[indices]["chain_id"], transl), axis=1)

    def distance(self, xyz=(0.00, 0.00, 0.00), records=("ATOM", "HETATM")):
        """Computes Euclidean distance between atoms and a 3D point.

        Parameters
        ----------
        xyz : tuple, default: (0.00, 0.00, 0.00)
            X, Y, and Z coordinate of the reference center for the distance
            computation.
        records : iterable, default: ('ATOM', 'HETATM')
            Specify which record sections to consider. For example, to consider
            both protein and ligand atoms, set `records=('ATOM', 'HETATM')`.
            This setting is ignored if `df` is not set to None.
            For downward compatibility, a string argument is still supported
            but deprecated and will be removed in future versions.

        Returns
        ---------
        pandas.Series : Pandas Series object containing the Euclidean
            distance between the atoms in the record section and `xyz`.

        """

        if isinstance(records, str):
            warnings.warn(
                "Using a string as `records` argument is "
                "deprecated and will not be supported in future"
                " versions. Please use a tuple or"
                " other iterable instead",
                DeprecationWarning,
            )
            records = (records,)

        df = pd.concat(objs=[self.df[i] for i in records])

        return np.sqrt(
            np.sum(
                df[["x_coord", "y_coord", "z_coord"]].subtract(xyz, axis=1) ** 2, axis=1
            )
        )

    @staticmethod
    def distance_df(df, xyz=(0.00, 0.00, 0.00)):
        """Computes Euclidean distance between atoms and a 3D point.

        Parameters
        ----------
        df : DataFrame
            DataFrame containing entries in the `PandasPdb.df['ATOM']`
            or `PandasPdb.df['HETATM']` format for the
            the distance computation to the `xyz` reference coordinates.
        xyz : tuple, default: (0.00, 0.00, 0.00)
            X, Y, and Z coordinate of the reference center for the distance
            computation.

        Returns
        ---------
        pandas.Series : Pandas Series object containing the Euclidean
            distance between the atoms in the record section and `xyz`.

        """
        return np.sqrt(
            np.sum(
                df[["x_coord", "y_coord", "z_coord"]].subtract(xyz, axis=1) ** 2, axis=1
            )
        )

    def to_pdb(self, path, records=None, gz=False, append_newline=True):
        """Write record DataFrames to a PDB file or gzipped PDB file.

        Parameters
        ----------
        path : str
            A valid output path for the pdb file

        records : iterable, default: None
            A list of PDB record sections in
            {'ATOM', 'HETATM', 'ANISOU', 'OTHERS'} that are to be written.
            Writes all lines to PDB if `records=None`.

        gz : bool, default: False
            Writes a gzipped PDB file if True.

        append_newline : bool, default: True
            Appends a new line at the end of the PDB file if True

        """
        if gz:
            openf = gzip.open
            w_mode = "wt"
        else:
            openf = open
            w_mode = "w"
        if not records:
            records = self.df.keys()

        dfs = {r: self.df[r].copy() for r in records if not self.df[r].empty}

        for r in dfs:
            for col in pdb_records[r]:
                dfs[r][col["id"]] = dfs[r][col["id"]].apply(col["strf"])
                dfs[r]["OUT"] = pd.Series("", index=dfs[r].index)

            for c in dfs[r].columns:
                # fix issue where coordinates with four or more digits would
                # cause issues because the columns become too wide
                if c in {"x_coord", "y_coord", "z_coord"}:
                    for idx in range(dfs[r][c].values.shape[0]):
                        if len(dfs[r][c].values[idx]) > 8:
                            dfs[r][c].values[idx] = str(dfs[r][c].values[idx]).strip()
                if c in {"line_idx", "OUT"}:
                    pass
                elif r in {"ATOM", "HETATM"} and c not in pdb_df_columns:
                    warn(
                        "Column %s is not an expected column and"
                        " will be skipped." % c
                    )
                else:
                    dfs[r]["OUT"] = dfs[r]["OUT"] + dfs[r][c]

        if pd_version < LooseVersion("0.17.0"):
            warn(
                "You are using an old pandas version (< 0.17)"
                " that relies on the old sorting syntax."
                " Please consider updating your pandas"
                " installation to a more recent version.",
                DeprecationWarning,
            )
            dfs.sort(columns="line_idx", inplace=True)

        elif pd_version < LooseVersion("0.23.0"):
            df = pd.concat(dfs)

        else:
            df = pd.concat(dfs, sort=False)

        df.sort_values(by="line_idx", inplace=True)

        with openf(path, w_mode) as f:

            s = df["OUT"].tolist()
            for idx in range(len(s)):
                if len(s[idx]) < 80:
                    s[idx] = f"{s[idx]}{' ' * (80 - len(s[idx]))}"
            to_write = "\n".join(s)
            f.write(to_write)
            if append_newline:
                f.write("\n")

    def parse_sse(self):
        """Parse secondary structure elements"""
        raise NotImplementedError

    def get_model_start_end(self) -> pd.DataFrame:
        """Get the start and end of the models contained in the PDB file.

        Extracts model start and end line indexes based
          on lines labelled 'OTHERS' during parsing.

        Returns
        ---------
        pandas.DataFrame : Pandas DataFrame object containing
          the start and end line indexes of the models.
        """

        other_records = self.df["OTHERS"]

        idxs = other_records.loc[other_records["record_name"] == "MODEL"].copy()
        ends = other_records.loc[other_records["record_name"] == "ENDMDL"]
        idxs.columns = ["record_name", "model_idx", "start_idx"]
        idxs.loc[:, "end_idx"] = ends.line_idx.values
        # If structure only contains 1 model, create a dummy df mapping all lines to model_idx 1
        if len(idxs) == 0:
            n_lines = len(self.pdb_text.splitlines())
            idxs = pd.DataFrame(
                [
                    {
                        "record_name": "MODEL",
                        "model_idx": 1,
                        "start_idx": 0,
                        "end_idx": n_lines,
                    }
                ]
            )

        return idxs

    def label_models(self):
        """Adds a column (`"model_id"`) to the underlying
        DataFrames containing the model number."""
        idxs = self.get_model_start_end()
        # Label ATOMS
        if "ATOM" in self.df.keys():
            pdb_df = self.df["ATOM"]
            idx_map = np.piecewise(
                np.zeros(len(pdb_df)),
                [
                    (pdb_df.line_idx.values >= start_idx)
                    & (pdb_df.line_idx.values <= end_idx)
                    for start_idx, end_idx in zip(
                        idxs.start_idx.values, idxs.end_idx.values
                    )
                ],
                idxs.model_idx,
            )
            self.df["ATOM"]["model_id"] = idx_map.astype(int)
        # LABEL HETATMS
        if "HETATM" in self.df.keys():
            pdb_df = self.df["HETATM"]
            idx_map = np.piecewise(
                np.zeros(len(pdb_df)),
                [
                    (pdb_df.line_idx.values >= start_idx)
                    & (pdb_df.line_idx.values <= end_idx)
                    for start_idx, end_idx in zip(
                        idxs.start_idx.values, idxs.end_idx.values
                    )
                ],
                idxs.model_idx,
            )
            self.df["HETATM"]["model_id"] = idx_map.astype(int)
        if "ANISOU" in self.df.keys():
            pdb_df = self.df["ANISOU"]
            idx_map = np.piecewise(
                np.zeros(len(pdb_df)),
                [
                    (pdb_df.line_idx.values >= start_idx)
                    & (pdb_df.line_idx.values <= end_idx)
                    for start_idx, end_idx in zip(
                        idxs.start_idx.values, idxs.end_idx.values
                    )
                ],
                idxs.model_idx,
            )
            self.df["ANISOU"]["model_id"] = idx_map.astype(int)
        return self

    def get_model(self, model_index: int) -> PandasPdb:
        """Returns a new PandasPDB object with the dataframes subset to the given model index.

        Parameters
        ----------
        model_index : int
            An integer representing the model index to subset to.

        Returns
        ---------
        pandas_pdb.PandasPdb : A new PandasPdb object containing the
          structure subsetted to the given model.
        """

        df = deepcopy(self)
        df.label_models()

        if "ATOM" in df.df.keys():
            df.df["ATOM"] = df.df["ATOM"].loc[df.df["ATOM"]["model_id"] == model_index]
        if "HETATM" in df.df.keys():
            df.df["HETATM"] = df.df["HETATM"].loc[
                df.df["HETATM"]["model_id"] == model_index
            ]
        if "ANISOU" in df.df.keys():
            df.df["ANISOU"] = df.df["ANISOU"].loc[
                df.df["ANISOU"]["model_id"] == model_index
            ]
        return df

    def get_models(self, model_indices: List[int]) -> PandasPdb:
        """Returns a new PandasPDB object with the dataframes subset to the given model index.

        Parameters
        ----------
        model_indices : List[int]
            A list representing the model indexes to subset to.

        Returns
        ---------
        pandas_pdb.PandasPdb : A new PandasPdb object
          containing the structure subsetted to the given model.
        """

        df = deepcopy(self)
        df.label_models()

        if "ATOM" in df.df.keys():
            df.df["ATOM"] = df.df["ATOM"].loc[
                [x in model_indices for x in df.df["ATOM"]["model_id"].tolist()]
            ]
        if "HETATM" in df.df.keys():
            df.df["HETATM"] = df.df["HETATM"].loc[
                [x in model_indices for x in df.df["HETATM"]["model_id"].tolist()]
            ]
        if "ANISOU" in df.df.keys():
            df.df["ANISOU"] = df.df["ANISOU"].loc[
                [x in model_indices for x in df.df["ANISOU"]["model_id"].tolist()]
            ]
        return df

    def to_pdb_stream(self, records: tuple[str] = ("ATOM", "HETATM")) -> StringIO:
        """Writes a PDB dataframe to a stream.

        Parameters
        ------------
        records : iterable, default: ('ATOM', 'HETATM')
            Iterable of record names to save to stream. Any of `["ATOM", "HETATM", "OTHERS"]`.

        Returns
        --------
        io.StringIO : Filestream of PDB file.
        """

        df = self.df.copy()
        df = pd.concat([df[a] for a in records])
        if "model_id" in df.columns:
            df = df.drop(columns=["model_id"])
        df.residue_number = df.residue_number.astype(int)
        records = [r.strip() for r in list(set(df.record_name))]
        dfs = {r: df.loc[df.record_name == r] for r in records}

        for r in dfs:
            for col in pdb_records[r]:
                dfs[r][col["id"]] = dfs[r][col["id"]].apply(col["strf"])
                dfs[r]["OUT"] = pd.Series("", index=dfs[r].index)

            for c in dfs[r].columns:
                # fix issue where coordinates with four or more digits would
                # cause issues because the columns become too wide
                if c in {"x_coord", "y_coord", "z_coord"}:
                    for idx in range(dfs[r][c].values.shape[0]):
                        if len(dfs[r][c].values[idx]) > 8:
                            dfs[r][c].values[idx] = str(
                                dfs[r][c].values[idx]).strip()

                if c not in {"line_idx", "OUT"}:
                    dfs[r]["OUT"] = dfs[r]["OUT"] + dfs[r][c]

        df = pd.concat(dfs, sort=False)
        df.sort_values(by="line_idx", inplace=True)

        output = StringIO()
        s = df["OUT"].tolist()
        for idx in range(len(s)):
            if len(s[idx]) < 80:
                s[idx] = f"{s[idx]}{' ' * (80 - len(s[idx]))}"
        to_write = "\n".join(s)
        output.write(to_write)
        output.write("\n")
        output.seek(0)
        return output

    def gyradius(self, records: tuple[str] = ("ATOM",),  decimals: int = 4) -> float:
        """Compute the Radius of Gyration of a molecule

        Parameters
        ----------
        records : iterable, default: ("ATOM",)
            Records from PandasPdb object for which to calculate the radius of gyration.
            Any of `("ATOM", "HETATM")`.

        decimals : int, default: 4
            Specifies the number of decimal places to round the final value to.

        Returns
        ---------
        rg : float
            Radius of Gyration of df in Angstrom

            """
        if isinstance(records, str):
            warnings.warn(
                "Using a string as `records` argument is "
                "deprecated and will not be supported in future "
                "versions. Please use a tuple or "
                "other iterable instead",
                DeprecationWarning,
            )
            records = (records,)

        if len(records) > 1:
            df = pd.concat(objs=[self.df[record][["x_coord",
                                                  "y_coord",
                                                  "z_coord",
                                                  "element_symbol"]]
                                 for record in records])
        else:
            df = self.df[records[0]]

        coords = df[["x_coord", "y_coord", "z_coord"]].to_numpy()
        masses = df["element_symbol"].map(lambda atom: ATOMIC_MASSES.get(atom, 0)).to_numpy()
        total_mass = masses.sum()
        center_of_mass = (masses[:, None] * coords).sum(axis=0) / total_mass
        distances = np.linalg.norm(coords - center_of_mass, axis=1)
        rg = np.sqrt((distances**2 * masses).sum() / total_mass)
        return round(rg, decimals)



# BioPandas
# Authors: Sebastian Raschka <mail@sebastianraschka.com>
# Authors: Arian Jamasb <arian@jamasb.io>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas

"""
BioPandas module for working with mmCIF
files in pandas DataFrames.
"""

from .pandas_mmcif import PandasMmcif

__all__ = ["PandasMmcif"]



from typing import Any, Dict, List

mmcif_col_types: Dict[str, Any] = {
    "B_iso_or_equiv": float,
    "Cartn_x": float,
    "Cartn_y": float,
    "Cartn_z": float,
    "auth_asym_id": str,
    "auth_atom_id": str,
    "auth_comp_id": str,
    "auth_seq_id": int,
    "group_PDB": str,
    "id": int,
    "label_alt_id": int,
    "label_asym_id": str,
    "label_atom_id": str,
    "label_comp_id": str,
    "label_entity_id": int,
    "label_seq_id": int,
    "occupancy": float,
    "pdbx_PDB_ins_code": str,
    "pdbx_PDB_model_num": int,
    "pdbx_formal_charge": int,
    "type_symbol": str,
}

ANISOU_DF_COLUMNS: List[str] = [
    "id",
    "type_symbol",
    "pdbx_label_atom_id",
    "pdbx_label_alt_id",
    "pdbx_label_comp_id",
    "pdbx_label_asym_id",
    "pdbx_label_seq_id",
    "pdbx_PDB_ins_code",
    "U[1][1]",
    "U[2][2]",
    "U[3][3]",
    "U[1][2]",
    "U[1][3]",
    "U[2][3]",
    "pdbx_auth_seq_id",
    "pdbx_auth_comp_id",
    "pdbx_auth_asym_id",
    "pdbx_auth_atom_id",
]


PDB_COLUMN_ORDER: List[str] = [
    "record_name",
    "atom_number",
    "blank_1",
    "atom_name",
    "alt_loc",
    "residue_name",
    "blank_2",
    "chain_id",
    "residue_number",
    "insertion",
    "blank_3",
    "x_coord",
    "y_coord",
    "z_coord",
    "occupancy",
    "b_factor",
    "blank_4",
    "segment_id",
    "element_symbol",
    "charge",
    "line_idx",
]

MMCIF_PDB_COLUMN_MAP: Dict[str, str] = {
    "group_PDB": "record_name",
    "id": "atom_number",
    "auth_atom_id": "atom_name",
    "auth_comp_id": "residue_name",
    "auth_asym_id": "chain_id",
    "auth_seq_id": "residue_number",
    "Cartn_x": "x_coord",
    "Cartn_y": "y_coord",
    "Cartn_z": "z_coord",
    "occupancy": "occupancy",
    "B_iso_or_equiv": "b_factor",
    "type_symbol": "element_symbol",
}

MMCIF_PDB_NONEFIELDS: List[str] = [
    "blank_1",
    "alt_loc",
    "blank_2",
    "insertion",
    "blank_3",
    "blank_4",
    "segment_id",
    "charge",
    "line_idx",
]



# BioPandas
# Author: Sebastian Raschka <mail@sebastianraschka.com>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas

import pandas as pd

amino3to1dict = {
    "ASH": "A",
    "ALA": "A",
    "CYX": "C",
    "CYS": "C",
    "ASP": "D",
    "GLU": "E",
    "PHE": "F",
    "GLY": "G",
    "HIS": "H",
    "HID": "H",
    "HIE": "H",
    "HIP": "H",
    "ILE": "I",
    "LYS": "K",
    "LEU": "L",
    "MET": "M",
    "MSE": "M",
    "ASN": "N",
    "PYL": "O",
    "HYP": "P",
    "PRO": "P",
    "GLN": "Q",
    "ARG": "R",
    "SER": "S",
    "THR": "T",
    "SEL": "U",
    "VAL": "V",
    "TRP": "W",
    "TYR": "Y",
}

pdb_df_columns = {
    "record_name",
    "atom_number",
    "blank_1",
    "atom_name",
    "alt_loc",
    "residue_name",
    "blank_2",
    "chain_id",
    "residue_number",
    "insertion",
    "blank_3",
    "x_coord",
    "y_coord",
    "z_coord",
    "occupancy",
    "b_factor",
    "blank_4",
    "segment_id",
    "element_symbol",
    "charge",
}

pdb_atomdict = [
    {"id": "record_name", "line": [0, 6], "type": str, "strf": lambda x: "%-6s" % x},
    {
        "id": "atom_number",
        "line": [6, 11],
        "type": int,
        "strf": lambda x: "%+5s" % str(x),
    },
    {"id": "blank_1", "line": [11, 12], "type": str, "strf": lambda x: "%-1s" % x},
    {
        "id": "atom_name",
        "line": [12, 16],
        "type": str,
        "strf": lambda x: " %-3s" % x if len(x) < 4 else "%-4s" % x,
    },
    {"id": "alt_loc", "line": [16, 17], "type": str, "strf": lambda x: "%-1s" % x},
    {"id": "residue_name", "line": [17, 20], "type": str, "strf": lambda x: "%+3s" % x},
    {"id": "blank_2", "line": [20, 21], "type": str, "strf": lambda x: "%-1s" % x},
    {"id": "chain_id", "line": [21, 22], "type": str, "strf": lambda x: "%-1s" % x},
    {
        "id": "residue_number",
        "line": [22, 26],
        "type": int,
        "strf": lambda x: "%+4s" % str(x),
    },
    {"id": "insertion", "line": [26, 27], "type": str, "strf": lambda x: "%-1s" % x},
    {"id": "blank_3", "line": [27, 30], "type": str, "strf": lambda x: "%-3s" % x},
    {
        "id": "x_coord",
        "line": [30, 38],
        "type": float,
        "strf": lambda x: ("%+8.3f" % x).replace("+", " "),
    },
    {
        "id": "y_coord",
        "line": [38, 46],
        "type": float,
        "strf": lambda x: ("%+8.3f" % x).replace("+", " "),
    },
    {
        "id": "z_coord",
        "line": [46, 54],
        "type": float,
        "strf": lambda x: ("%+8.3f" % x).replace("+", " "),
    },
    {
        "id": "occupancy",
        "line": [54, 60],
        "type": float,
        "strf": lambda x: ("%+6.2f" % x).replace("+", " "),
    },
    {
        "id": "b_factor",
        "line": [60, 66],
        "type": float,
        "strf": lambda x: ("%+6.2f" % x).replace("+", " ") if len(str(int(x))) < 3 else ("%+6.2f" % x).replace("+", ""),
    },
    {"id": "blank_4", "line": [66, 72], "type": str, "strf": lambda x: "%-7s" % x},
    {"id": "segment_id", "line": [72, 76], "type": str, "strf": lambda x: "%-3s" % x},
    {
        "id": "element_symbol",
        "line": [76, 78],
        "type": str,
        "strf": lambda x: "%+2s" % x,
    },
    {
        "id": "charge",
        "line": [78, 80],
        "type": float,
        "strf": lambda x: (("%+2.1f" % x).replace("+", " ") if pd.notnull(x) else ""),
    },
]


pdb_anisoudict = [
    {"id": "record_name", "line": [0, 6], "type": str, "strf": lambda x: "%-6s" % x},
    {
        "id": "atom_number",
        "line": [6, 11],
        "type": int,
        "strf": lambda x: "%+5s" % str(x),
    },
    {"id": "blank_1", "line": [11, 12], "type": str, "strf": lambda x: "%-1s" % x},
    {
        "id": "atom_name",
        "line": [12, 16],
        "type": str,
        "strf": lambda x: (" %-3s" % x if len(x) < 4 else "%-4s" % x),
    },
    {"id": "alt_loc", "line": [16, 17], "type": str, "strf": lambda x: "%-1s" % x},
    {"id": "residue_name", "line": [17, 20], "type": str, "strf": lambda x: "%+3s" % x},
    {"id": "blank_2", "line": [20, 21], "type": str, "strf": lambda x: "%-1s" % x},
    {"id": "chain_id", "line": [21, 22], "type": str, "strf": lambda x: "%-1s" % x},
    {
        "id": "residue_number",
        "line": [22, 26],
        "type": int,
        "strf": lambda x: "%+4s" % str(x),
    },
    {"id": "insertion", "line": [26, 27], "type": str, "strf": lambda x: "%-1s" % x},
    {"id": "blank_3", "line": [27, 28], "type": str, "strf": lambda x: "%-1s" % x},
    {"id": "U(1,1)", "line": [28, 35], "type": int, "strf": lambda x: "%+7s" % str(x)},
    {"id": "U(2,2)", "line": [35, 42], "type": int, "strf": lambda x: "%+7s" % str(x)},
    {"id": "U(3,3)", "line": [42, 49], "type": int, "strf": lambda x: "%+7s" % str(x)},
    {"id": "U(1,2)", "line": [49, 56], "type": int, "strf": lambda x: "%+7s" % str(x)},
    {"id": "U(1,3)", "line": [56, 63], "type": int, "strf": lambda x: "%+7s" % str(x)},
    {"id": "U(2,3)", "line": [63, 70], "type": int, "strf": lambda x: "%+7s" % str(x)},
    {"id": "blank_4", "line": [70, 76], "type": str, "strf": lambda x: "%+6s" % x},
    {
        "id": "element_symbol",
        "line": [76, 78],
        "type": str,
        "strf": lambda x: "%+2s" % x,
    },
    {
        "id": "charge",
        "line": [78, 80],
        "type": float,
        "strf": lambda x: (("%+2.1f" % x).replace("+", " ") if pd.notnull(x) else ""),
    },
]

pdb_otherdict = [
    {
        "id": "record_name",
        "line": [0, 6],
        "type": str,
        "strf": lambda x: "%s%s" % (x, " " * (6 - len(x))),
    },
    {"id": "entry", "line": [6, -2], "type": str, "strf": lambda x: x.rstrip()},
]

pdb_records = {
    "ATOM": pdb_atomdict,
    "HETATM": pdb_atomdict,
    "ANISOU": pdb_anisoudict,
    "OTHERS": pdb_otherdict,
}



# BioPandas
# Author: Sebastian Raschka <mail@sebastianraschka.com>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas

import pandas as pd
from biopandas.pdb import PandasPdb
import os


def test_equal():
    TESTDATA_1t48 = os.path.join(os.path.dirname(__file__), "data", "1t48_995.pdb")

    p1t48 = PandasPdb()
    p1t48.read_pdb(TESTDATA_1t48)
    dist = p1t48.distance(xyz=(70.785, 15.477, 23.359), records=("ATOM",))

    expect = pd.Series(
        [2.533259, 1.520502, 0.000000, 1.257597, 1.252510], index=[12, 13, 14, 15, 16]
    )
    assert dist[dist < 3].all() == expect.all()


def test_deprecated_str_arg():
    TESTDATA_1t48 = os.path.join(os.path.dirname(__file__), "data", "1t48_995.pdb")

    p1t48 = PandasPdb()
    p1t48.read_pdb(TESTDATA_1t48)
    dist = p1t48.distance(xyz=(70.785, 15.477, 23.359), records="ATOM")

    expect = pd.Series(
        [2.533259, 1.520502, 0.000000, 1.257597, 1.252510], index=[12, 13, 14, 15, 16]
    )
    assert dist[dist < 3].all() == expect.all()


def test_use_external_df():
    TESTDATA_1t48 = os.path.join(os.path.dirname(__file__), "data", "1t48_995.pdb")

    p1t48 = PandasPdb()
    p1t48.read_pdb(TESTDATA_1t48)
    new_df = p1t48.df["ATOM"].iloc[:-1, :].copy()
    dist = PandasPdb.distance_df(df=new_df, xyz=(70.785, 15.477, 23.359))

    expect = pd.Series([2.533259, 1.520502, 0.000000, 1.257597], index=[12, 13, 14, 15])
    assert dist[dist < 3].all() == expect.all()



# BioPandas
# Author: Sebastian Raschka <mail@sebastianraschka.com>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas


import os
from urllib.error import HTTPError, URLError
from urllib.request import urlopen

import numpy as np
import pandas as pd
from nose.tools import raises

from biopandas.mmtf import PandasMmtf
from biopandas.pdb import PandasPdb
from biopandas.testutils import assert_raises

MMTF_TESTDATA_FILENAME = os.path.join(os.path.dirname(__file__), "data", "3eiy.mmtf")
MMTF_TESTDATA_FILENAME_GZ = os.path.join(os.path.dirname(__file__), "data", "3eiy.mmtf.gz")

PDB_TESTDATA_FILENAME = os.path.join(os.path.dirname(__file__), "..", "..", "pdb", "tests", "data", "3eiy.pdb")
PDB_TESTDATA_FILENAME_GZ = os.path.join(os.path.dirname(__file__), "..", "..", "pdb", "tests", "data", "3eiy.pdb.gz")


ATOM_DF_COLUMNS = [
    "record_name",
    "atom_number",
    "atom_name",
    #"alt_loc",
    "residue_name",
    "chain_id",
    "residue_number",
    #"insertion",
    "x_coord",
    "y_coord",
    "z_coord",
    "occupancy",
    "b_factor",
    "element_symbol",
    #"charge",
]

def test_fetch_pdb():
    """Test fetch_pdb"""
    ppdb = PandasMmtf()
    ppdb.fetch_mmtf("3eiy")
    assert max(ppdb.df["ATOM"].residue_number) == 175


def test__read_mmtf():
    """Test public _read_pdb with gzip files"""
    pmmtf = PandasMmtf()
    ppdb = PandasPdb()
    pmmtf.read_mmtf(MMTF_TESTDATA_FILENAME)

    ppdb = ppdb.read_pdb(PDB_TESTDATA_FILENAME)

    pd.testing.assert_frame_equal(
        pmmtf.df["ATOM"][ATOM_DF_COLUMNS].reset_index(drop=True),
        ppdb.df["ATOM"][ATOM_DF_COLUMNS].reset_index(drop=True),
        )

    ATOM_DF_COLUMNS.remove("atom_number")
    ATOM_DF_COLUMNS.remove("element_symbol")
    pd.testing.assert_frame_equal(
        pmmtf.df["HETATM"][ATOM_DF_COLUMNS].reset_index(drop=True),
        ppdb.df["HETATM"][ATOM_DF_COLUMNS].reset_index(drop=True),
        )


def test__read_mmtf_gz():
    """Test public _read_pdb with gzip files"""
    pmmtf = PandasMmtf()
    ppdb = PandasPdb()
    pmmtf.read_mmtf(MMTF_TESTDATA_FILENAME_GZ)
    ppdb = ppdb.read_pdb(PDB_TESTDATA_FILENAME_GZ)


    pmmtf.df["ATOM"].alt_loc.replace('\x00', "", inplace=True)
    pmmtf.df["HETATM"].alt_loc.replace('\x00', "", inplace=True)

    pd.testing.assert_frame_equal(
        pmmtf.df["ATOM"][ATOM_DF_COLUMNS].reset_index(drop=True),
        ppdb.df["ATOM"][ATOM_DF_COLUMNS].reset_index(drop=True),
        )
    #pd.testing.assert_frame_equal(
    #    pmmtf.df["HETATM"][ATOM_DF_COLUMNS].reset_index(drop=True),
    #    ppdb.df["HETATM"][ATOM_DF_COLUMNS].reset_index(drop=True),
    #    )


def test_read_mmtf():
    """Test public read_pdb"""
    ppdb = PandasMmtf()
    ppdb.read_mmtf(MMTF_TESTDATA_FILENAME)
    assert ppdb.mmtf_path == MMTF_TESTDATA_FILENAME







"""Class for working with MMCIF files."""
# BioPandas
# Authors: Arian Jamasb <arian@jamasb.io>,
# Authors: Sebastian Raschka <mail@sebastianraschka.com>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas

import gzip
import sys
import warnings
from typing import Dict, List, Optional
from urllib.error import HTTPError, URLError
from urllib.request import urlopen

import numpy as np
import pandas as pd
from looseversion import LooseVersion

from ..pdb.engines import amino3to1dict
from ..pdb.pandas_pdb import PandasPdb
from .engines import (ANISOU_DF_COLUMNS, MMCIF_PDB_COLUMN_MAP,
                      MMCIF_PDB_NONEFIELDS, PDB_COLUMN_ORDER, mmcif_col_types)
from .mmcif_parser import load_cif_data

pd_version = LooseVersion(pd.__version__)


class PandasMmcif:
    def __init__(self, use_auth: bool = True):
        self._df = None
        self.mmcif_text = ""
        self.header = ""
        self.code = ""
        self.mmcif_path = ""
        self.auth = use_auth
        self._get_dict = {}

    @property
    def df(self):
        """Acccess dictionary of pandas DataFrames for PDB record sections."""
        return self._df

    @df.setter
    def df(self, value):
        """Assign a new value to the pandas DataFrame"""
        raise AttributeError(
            "Please use `PandasMmcif._df = ... ` instead\n"
            "of `PandasMmcif.df = ... ` if you are sure that\n"
            "you want to overwrite the `df` attribute."
        )

    def read_mmcif(self, path):
        """Read MMCIF files (unzipped or gzipped) from local drive

        Attributes
        ----------
        path : Union[str, os.PathLike]
            Path to the MMCIF file in .cif format or gzipped format (.cif.gz).

        Returns
        ---------
        self

        """
        self.mmcif_path, self.pdb_text = self._read_mmcif(path=str(path))
        self._df = self._construct_df(text=self.pdb_text)
        # self.header, self.code = self._parse_header_code() #TODO: implement
        self.code = self.data["entry"]["id"][0].lower()
        return self

    def fetch_mmcif(self, pdb_code: Optional[str] = None, uniprot_id: Optional[str] = None, source: str = "pdb"):
        """Fetches mmCIF file contents from the Protein Databank at rcsb.org or AlphaFold database at https://alphafold.ebi.ac.uk/.
.

        Parameters
        ----------
        pdb_code : str, optional
            A 4-letter PDB code, e.g., `"3eiy"` to retrieve structures from the PDB. Defaults to `None`.

        uniprot_id : str, optional
            A UniProt Identifier, e.g., `"Q5VSL9"` to retrieve structures from the AF2 database. Defaults to `None`.

        source : str
            The source to retrieve the structure from 
            (`"pdb"`, `"alphafold2-v3"` or `"alphafold2-v4"`). Defaults to `"pdb"`.

        Returns
        ---------
        self

        """
        # Sanitize input
        invalid_input_identifier_1 = pdb_code is None and uniprot_id is None
        invalid_input_identifier_2 = pdb_code is not None and uniprot_id is not None
        invalid_input_combination_1 = uniprot_id is not None and source == "pdb"
        invalid_input_combination_2 = pdb_code is not None and source in {
            "alphafold2-v3", "alphafold2-v4"}

        if invalid_input_identifier_1 or invalid_input_identifier_2:
            raise ValueError(
                "Please provide either a PDB code or a UniProt ID.")

        if invalid_input_combination_1:
            raise ValueError(
                "Please use a 'pdb_code' instead of 'uniprot_id' for source='pdb'.")
        elif invalid_input_combination_2:
            raise ValueError(
                f"Please use a 'uniprot_id' instead of 'pdb_code' for source={source}.")

        if source == "pdb":
            self.mmcif_path, self.mmcif_text = self._fetch_mmcif(pdb_code)
        elif source == "alphafold2-v3":
            af2_version = 3
            self.mmcif_path, self.mmcif_text = self._fetch_af2(uniprot_id, af2_version)
        elif source == "alphafold2-v4":
            af2_version = 4
            self.mmcif_path, self.mmcif_text = self._fetch_af2(uniprot_id, af2_version)
        else:
            raise ValueError(f"Invalid source: {source}."
                " Please use one of 'pdb', 'alphafold2-v3' or 'alphafold2-v4'.")

        self._df = self._construct_df(text=self.mmcif_text)
        return self

    def _construct_df(self, text: str):
        data = load_cif_data(text)
        data = data[list(data.keys())[0]]
        self.data = data
        df: Dict[str, pd.DataFrame] = {}
        full_df = pd.DataFrame.from_dict(
            data["atom_site"], orient="index").transpose()
        full_df = full_df.astype(mmcif_col_types, errors="ignore")
        df["ATOM"] = pd.DataFrame(full_df[full_df.group_PDB == "ATOM"])
        df["HETATM"] = pd.DataFrame(full_df[full_df.group_PDB == "HETATM"])
        try:
            df["ANISOU"] = pd.DataFrame(data["atom_site_anisotrop"])
        except KeyError:
            df["ANISOU"] = pd.DataFrame(columns=ANISOU_DF_COLUMNS)
        return df

    @staticmethod
    def _fetch_mmcif(pdb_code):
        """Load MMCIF file from rcsb.org."""
        txt = None
        url = f"https://files.rcsb.org/download/{pdb_code.lower()}.cif"
        try:
            response = urlopen(url)
            txt = response.read()
            txt = (
                txt.decode(
                    "utf-8") if sys.version_info[0] >= 3 else txt.encode("ascii")
            )
        except HTTPError as e:
            print(f"HTTP Error {e.code}")
        except URLError as e:
            print(f"URL Error {e.args}")
        return url, txt

    @staticmethod
    def _fetch_af2(uniprot_id: str, af2_version: int = 3):
        """Load MMCIF file from https://alphafold.ebi.ac.uk/."""
        txt = None
        url = f"https://alphafold.ebi.ac.uk/files/AF-{uniprot_id.upper()}-F1-model_v{af2_version}.cif"

        try:
            response = urlopen(url)
            txt = response.read()
            txt = txt.decode('utf-8') if sys.version_info[0] >= 3 else txt.encode('ascii')
        except HTTPError as e:
            print(f'HTTP Error {e.code}')
        except URLError as e:
            print(f'URL Error {e.args}')
        return url, txt

    @staticmethod
    def _read_mmcif(path):
        """Read MMCIF file from local drive."""
        r_mode = "r"
        if path.endswith((".cif", ".mmcif")):
            openf = open
        elif path.endswith((".cif.gz", ".mmcif.gz")):
            r_mode = "rb"
            openf = gzip.open
        else:
            allowed_formats = ", ".join(
                (".cif", ".cif.gz", ".mmcif", ".mmcif.gz"))
            raise ValueError(
                f"Wrong file format; allowed file formats are {allowed_formats}"
            )

        with openf(path, r_mode) as f:
            txt = f.read()

        if path.endswith(".gz"):
            txt = (
                txt.decode(
                    "utf-8") if sys.version_info[0] >= 3 else txt.encode("ascii")
            )
        return path, txt

    def get(self, s, df=None, invert=False, records=("ATOM", "HETATM")):
        """Filter PDB DataFrames by properties

        Parameters
        ----------
        s : str  in {'main chain', 'hydrogen', 'c-alpha', 'heavy'}
            String to specify which entries to return.

        df : pandas.DataFrame, default: None
            Optional DataFrame to perform the filter operation on.
            If df=None, filters on self.df['ATOM'].

        invert : bool, default: True
            Inverts the search query. For example if s='hydrogen' and
            invert=True, all but hydrogen entries are returned.

        records : iterable, default: ('ATOM', 'HETATM')
            Specify which record sections to consider. For example, to consider
            both protein and ligand atoms, set `records=('ATOM', 'HETATM')`.
            This setting is ignored if `df` is not set to None.
            For downward compatibility, a string argument is still supported
            but deprecated and will be removed in future versions.

        Returns
        --------
        df : pandas.DataFrame
            Returns a DataFrame view on the filtered entries.

        """
        if isinstance(records, str):
            warnings.warn(
                "Using a string as `records` argument is "
                "deprecated and will not be supported in future"
                " versions. Please use a tuple or"
                " other iterable instead",
                DeprecationWarning,
            )
            records = (records,)

        if not self._get_dict:
            self._get_dict = self._init_get_dict()
        if s not in self._get_dict.keys():
            raise AttributeError(f"s must be in {self._get_dict.keys()}")
        if not df:
            df = pd.concat(objs=[self.df[i] for i in records])
        return self._get_dict[s](df, invert=invert)

    @staticmethod
    def _get_mainchain(
        df: pd.DataFrame, invert: bool = False, atom_col: str = "auth_atom_id"
    ) -> pd.DataFrame:
        """Return only main chain atom entries from a DataFrame"""
        return (
            df[
                (df[atom_col] != "C")
                & (df[atom_col] != "O")
                & (df[atom_col] != "N")
                & (df[atom_col] != "CA")
            ]
            if invert
            else df[
                (df[atom_col] == "C")
                | (df[atom_col] == "O")
                | (df[atom_col] == "N")
                | (df[atom_col] == "CA")
            ]
        )

    @staticmethod
    def _get_hydrogen(df, invert):
        """Return only hydrogen atom entries from a DataFrame"""
        return (
            df[(df["type_symbol"] != "H")] if invert else df[(
                df["type_symbol"] == "H")]
        )

    @staticmethod
    def _get_heavy(df, invert):
        """Return only heavy atom entries from a DataFrame"""
        return df[df["type_symbol"] == "H"] if invert else df[df["type_symbol"] != "H"]

    @staticmethod
    def _get_calpha(df, invert, atom_col: str = "auth_atom_id"):
        """Return c-alpha atom entries from a DataFrame"""
        return df[df[atom_col] != "CA"] if invert else df[df[atom_col] == "CA"]

    @staticmethod
    def _get_carbon(df, invert):
        """Return carbon atom entries from a DataFrame"""
        return df[df["type_symbol"] != "C"] if invert else df[df["type_symbol"] == "C"]

    def amino3to1(
        self,
        record: str = "ATOM",
        residue_col: str = "auth_comp_id",
        residue_number_col: str = "auth_seq_id",
        chain_col: str = "auth_asym_id",
        fillna: str = "?",
    ):
        """Creates 1-letter amino acid codes from DataFrame

        Non-canonical amino-acids are converted as follows:
        ASH (protonated ASP) => D
        CYX (disulfide-bonded CYS) => C
        GLH (protonated GLU) => E
        HID/HIE/HIP (different protonation states of HIS) = H
        HYP (hydroxyproline) => P
        MSE (selenomethionine) => M

        Parameters
        ----------
        record : str, default: 'ATOM'
            Specfies the record DataFrame.
        residue_col : str,  default: 'residue_name'
            Column in `record` DataFrame to look for 3-letter amino acid
            codes for the conversion.
        fillna : str, default: '?'
            Placeholder string to use for unknown amino acids.

        Returns
        ---------
        pandas.DataFrame : Pandas DataFrame object consisting of two columns,
            `'chain_id'` and `'residue_name'`, where the former contains
            the chain ID of the amino acid and the latter
            contains the 1-letter amino acid code, respectively.

        """
        tmp = self.df[record]
        cmp = "placeholder"
        indices = []

        residue_number_insertion = (
            tmp[residue_number_col].astype(str) + tmp["pdbx_PDB_ins_code"]
        )

        for num, ind in zip(residue_number_insertion, np.arange(tmp.shape[0])):
            if num != cmp:
                indices.append(ind)
            cmp = num

        transl = tmp.iloc[indices][residue_col].map(
            amino3to1dict).fillna(fillna)

        return pd.concat((tmp.iloc[indices][chain_col], transl), axis=1)

    @staticmethod
    def rmsd(df1, df2, s=None, invert=False):
        """Compute the Root Mean Square Deviation between molecules.

        Parameters
        ----------
        df1 : pandas.DataFrame
            DataFrame with HETATM, ATOM, and/or ANISOU entries.

        df2 : pandas.DataFrame
            Second DataFrame for RMSD computation against df1. Must have the
            same number of entries as df1.

        s : {'main chain', 'hydrogen', 'c-alpha', 'heavy', 'carbon'} or None,
            default: None
            String to specify which entries to consider. If None, considers
            all atoms for comparison.

        invert : bool, default: False
            Inverts the string query if true. For example, the setting
            `s='hydrogen', invert=True` computes the RMSD based on all
            but hydrogen atoms.

        Returns
        ---------
        rmsd : float
            Root Mean Square Deviation between df1 and df2

        """
        if df1.shape[0] != df2.shape[0]:
            raise AttributeError("DataFrames have unequal lengths")
        get_dict = PandasMmcif._init_get_dict()
        if s:
            if s not in get_dict.keys():
                raise AttributeError(f"s must be in {get_dict.keys()} or None")
            df1 = get_dict[s](df1, invert=invert)
            df2 = get_dict[s](df2, invert=invert)

        total = (
            (df1["Cartn_x"].values - df2["Cartn_x"].values) ** 2
            + (df1["Cartn_y"].values - df2["Cartn_y"].values) ** 2
            + (df1["Cartn_z"].values - df2["Cartn_z"].values) ** 2
        )
        return round((total.sum() / df1.shape[0]) ** 0.5, 4)

    def distance(self, xyz=(0.00, 0.00, 0.00), records=("ATOM", "HETATM")):
        """Computes Euclidean distance between atoms and a 3D point.

        Parameters
        ----------
        xyz : tuple, default: (0.00, 0.00, 0.00)
            X, Y, and Z coordinate of the reference center for the distance
            computation.
        records : iterable, default: ('ATOM', 'HETATM')
            Specify which record sections to consider. For example, to consider
            both protein and ligand atoms, set `records=('ATOM', 'HETATM')`.
            This setting is ignored if `df` is not set to None.
            For downward compatibility, a string argument is still supported
            but deprecated and will be removed in future versions.

        Returns
        ---------
        pandas.Series : Pandas Series object containing the Euclidean
            distance between the atoms in the record section and `xyz`.

        """

        if isinstance(records, str):
            warnings.warn(
                "Using a string as `records` argument is "
                "deprecated and will not be supported in future"
                " versions. Please use a tuple or"
                " other iterable instead",
                DeprecationWarning,
            )
            records = (records,)

        df = pd.concat(objs=[self.df[i] for i in records])

        return np.sqrt(
            np.sum(
                df[["Cartn_x", "Cartn_y", "Cartn_z"]].subtract(xyz, axis=1) ** 2, axis=1
            )
        )

    @staticmethod
    def distance_df(df, xyz=(0.00, 0.00, 0.00)):
        """Computes Euclidean distance between atoms and a 3D point.

        Parameters
        ----------
        df : DataFrame
            DataFrame containing entries in the `PandasPdb.df['ATOM']`
            or `PandasPdb.df['HETATM']` format for the
            the distance computation to the `xyz` reference coordinates.
        xyz : tuple, default: (0.00, 0.00, 0.00)
            X, Y, and Z coordinate of the reference center for the distance
            computation.

        Returns
        ---------
        pandas.Series : Pandas Series object containing the Euclidean
            distance between the atoms in the record section and `xyz`.

        """
        return np.sqrt(
            np.sum(
                df[["Cartn_x", "Cartn_y", "Cartn_z"]].subtract(xyz, axis=1) ** 2, axis=1
            )
        )

    @staticmethod
    def _init_get_dict():
        """Initialize dictionary for filter operations."""
        return {
            "main chain": PandasMmcif._get_mainchain,
            "hydrogen": PandasMmcif._get_hydrogen,
            "c-alpha": PandasMmcif._get_calpha,
            "carbon": PandasMmcif._get_carbon,
            "heavy": PandasMmcif._get_heavy,
        }

    def read_mmcif_from_list(self, mmcif_lines):
        """Reads mmCIF file from a list into DataFrames

        Attributes
        ----------
        pdb_lines : list
            A list of lines containing the mmCIF file contents.

        Returns
        ---------
        self

        """
        self.pdb_text = "".join(mmcif_lines)
        self._df = self._construct_df(mmcif_lines)
        # self.header, self.code = self._parse_header_code()
        self.code = self.data["entry"]["id"][0].lower()
        return self

    def convert_to_pandas_pdb(self, offset_chains: bool = True, records: List[str] = ["ATOM", "HETATM"]) -> PandasPdb:
        """Returns a PandasPdb object with the same data as the PandasMmcif
        object.

        Attributes
        ----------
        offset_chains: bool
            Whether or not to offset atom numbering based on number of chains.
            This can arise due to the presence of TER records in PDBs which are
            not found in mmCIFs.
        records: List[str]
            List of record types to save. Any of ["ATOM", "HETATM", "OTHERS"].
            Defaults to ["ATOM", "HETATM"].

        """
        pandaspdb = PandasPdb()

        for a in records:
            try:
                dfa = self.df[a]
                # keep only those fields found in pdb
                dfa = dfa[MMCIF_PDB_COLUMN_MAP.keys()]
                # rename fields
                dfa = dfa.rename(columns=MMCIF_PDB_COLUMN_MAP)
                # add empty fields
                for i in MMCIF_PDB_NONEFIELDS:
                    dfa[i] = ""
                dfa["charge"] = np.nan
                # reorder columns to PandasPdb order
                dfa = dfa[PDB_COLUMN_ORDER]
                pandaspdb.df[a] = dfa
            except KeyError:  # Some entries may not have an ANISOU
                continue

        # update line_idx
        pandaspdb.df["ATOM"]["line_idx"] = pandaspdb.df["ATOM"].index.values
        pandaspdb.df["HETATM"]["line_idx"] = pandaspdb.df["HETATM"].index

        # Update atom numbers
        if offset_chains:
            offsets = pandaspdb.df["ATOM"]["chain_id"].astype(
                "category").cat.codes
            pandaspdb.df["ATOM"]["atom_number"] = pandaspdb.df["ATOM"]["atom_number"] + offsets
            hetatom_offset = offsets.max() + 1
            pandaspdb.df["HETATM"]["atom_number"] = pandaspdb.df["HETATM"]["atom_number"] + hetatom_offset

        return pandaspdb



# BioPandas
# Authors: Sebastian Raschka <mail@sebastianraschka.com>
# Authors: Arian Jamasb <arian@jamasb.io>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas

"""
BioPandas module for working with MMTF
files in pandas DataFrames.
"""

from .pandas_mmtf import PandasMmtf, fetch_mmtf, parse_mmtf

__all__ = ["PandasMmtf", "fetch_mmtf", "parse_mmtf"]



# BioPandas
# Author: Sebastian Raschka <mail@sebastianraschka.com>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas


from biopandas.pdb import PandasPdb
import os
import numpy as np
import pandas as pd
from nose.tools import raises

TESTDATA_FILENAME = os.path.join(
    os.path.dirname(__file__), "data", "3eiy_stripped_no_ele.pdb"
)

ppdb = PandasPdb()
ppdb.read_pdb(TESTDATA_FILENAME)


def test_impute_hetatm():
    new = ppdb.impute_element(records=["HETATM"])
    assert new["HETATM"]["element_symbol"][1] == "N"
    assert new["HETATM"]["element_symbol"][10] == "O"
    assert new["ATOM"]["element_symbol"][1] == ""
    assert new["ATOM"]["element_symbol"][10] == ""


def test_impute_atom():
    new = ppdb.impute_element(records=["ATOM"])
    assert new["ATOM"]["element_symbol"][1] == "C"
    assert new["ATOM"]["element_symbol"][10] == "C"
    assert new["HETATM"]["element_symbol"][1] == ""
    assert new["HETATM"]["element_symbol"][10] == ""


def test_impute_default():
    assert ppdb.df["ATOM"]["element_symbol"][1] == ""
    assert ppdb.df["HETATM"]["element_symbol"][1] == ""
    new = ppdb.impute_element()
    assert new["ATOM"]["element_symbol"][1] == "C"
    assert new["ATOM"]["element_symbol"][10] == "C"
    assert new["HETATM"]["element_symbol"][1] == "N"
    assert new["HETATM"]["element_symbol"][10] == "O"



import os

import pandas as pd
from pandas.testing import assert_frame_equal

from biopandas.mmtf.pandas_mmtf import PandasMmtf, write_mmtf


def test_write_mmtf_bp():
    PDB_CODES = ["4hhb", "3eiy", "1t48", "1ehz", "4ggb", "1bxa", "1cbn", "1rcf"]
    for pdb in PDB_CODES:
        print(pdb)
        pm1 = PandasMmtf().fetch_mmtf(pdb)
        pm1.to_mmtf("test.mmtf")
        assert os.path.exists("test.mmtf")

        pm2 = PandasMmtf().read_mmtf("test.mmtf")
        assert_frame_equal(pm1.df["ATOM"].reset_index(drop=True), pm2.df["ATOM"].reset_index(drop=True))
        assert_frame_equal(pm1.df["HETATM"].reset_index(drop=True), pm2.df["HETATM"].reset_index(drop=True))

    os.remove("test.mmtf")

def test_write_mmtf():
    PDB_CODES = ["4hhb", "3eiy", "1t48", "1ehz", "4ggb", "1bxa", "1cbn", "1rcf"]
    for pdb in PDB_CODES:
        print(pdb)
        pm1 = PandasMmtf().fetch_mmtf(pdb)
        write_mmtf(pd.concat([pm1.df["ATOM"], pm1.df["HETATM"]]), "test.mmtf")
        assert os.path.exists("test.mmtf")

        pm2 = PandasMmtf().read_mmtf("test.mmtf")
        assert_frame_equal(pm1.df["ATOM"].reset_index(drop=True), pm2.df["ATOM"].reset_index(drop=True))
        assert_frame_equal(pm1.df["HETATM"].reset_index(drop=True), pm2.df["HETATM"].reset_index(drop=True))

    os.remove("test.mmtf")




# BioPandas
# License: BSD 3 clause
# Author goniochromatic: https://github.com/goniochromatic/
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas

from biopandas.pdb import PandasPdb
import os
import pandas as pd
from nose.tools import raises
import warnings

TESTDATA_1t48 = os.path.join(os.path.dirname(__file__), "data", "1t48_995.pdb")

p1t48 = PandasPdb()
p1t48.read_pdb(TESTDATA_1t48)


def test_default_record():
    rg = p1t48.gyradius()
    expected_rg = 18.1508
    assert rg == expected_rg, f"Expected {expected_rg}, got {rg} instead"


def test_atom():
    rg = p1t48.gyradius(("ATOM",))
    expected_rg = 18.1508
    assert rg == expected_rg, f"Expected {expected_rg}, got {rg} instead"


def test_hetatm():
    rg = p1t48.gyradius(("HETATM",))
    expected_rg = 21.0143
    assert rg == expected_rg, f"Expected {expected_rg}, got {rg} instead"


def test_atom_and_hetatm():
    rg = p1t48.gyradius(("ATOM", "HETATM"))
    expected_rg = 19.0307
    assert rg == expected_rg, f"Expected {expected_rg}, got {rg} instead"


@raises(KeyError)
def test_wrong_record_name():
    p1t48.gyradius(("Wrong",))


@raises(TypeError)
def test_wrong_arg_type():
    p1t48.gyradius(5)


def test_decimals():
    rg = p1t48.gyradius(decimals=5)
    expected_rg = 18.15084
    assert rg == expected_rg, f"Expected {expected_rg}, got {rg} instead"


def test_negative_decimals():
    rg = p1t48.gyradius(decimals=-1)
    expected_rg = 20.0
    assert rg == expected_rg, f"Expected {expected_rg}, got {rg} instead"


@raises(TypeError)
def test_wrong_decimals_arg():
    p1t48.gyradius(decimals='five')


def test_both_args():
    rg = p1t48.gyradius(("HETATM",), 5)
    expected_rg = 21.01426
    assert rg == expected_rg, f"Expected {expected_rg}, got {rg} instead"



# BioPandas
# Author: Sebastian Raschka <mail@sebastianraschka.com>
# Author: Arian Jamasb <arian@jamasb.io>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas
import os

from pandas.testing import assert_frame_equal

from biopandas.mmtf import PandasMmtf

TESTDATA_FILENAME = os.path.join(os.path.dirname(__file__), "data", "2jyf.mmtf")


def test_label_models():
    df = PandasMmtf().read_mmtf(TESTDATA_FILENAME)
    assert "model_id" in df.df["ATOM"].columns


def test_get_model():
    df = PandasMmtf().read_mmtf(TESTDATA_FILENAME)
    MODEL_INDEX = 1
    new_df = df.get_model(MODEL_INDEX)
    assert new_df.df["ATOM"]["model_id"].all() == MODEL_INDEX


def test_get_models():
    df = PandasMmtf().read_mmtf(TESTDATA_FILENAME)
    MODEL_INDICES = [1, 3, 5]

    df = df.get_models(MODEL_INDICES)
    assert df.df["ATOM"]["model_id"].all() in MODEL_INDICES

    # Test a round trip through biopandas
    df.to_mmtf("test.mmtf")
    written = PandasMmtf().read_mmtf("test.mmtf")

    # Note: No way to preserve model ID as far as I can tell
    assert_frame_equal(df.df["ATOM"].drop("model_id", axis=1).reset_index(drop=True), written.df["ATOM"].drop("model_id", axis=1).reset_index(drop=True))
    assert_frame_equal(df.df["HETATM"].drop("model_id", axis=1).reset_index(drop=True), written.df["HETATM"].drop("model_id", axis=1).reset_index(drop=True))

    # Clean
    os.remove("test.mmtf")



# the cif parser relieves heavily on the parser from pdb_japan
# (author: gert-jan bekker).
# python cif parser: https://gitlab.com/pdbjapan/tools/cif-parsers
# license: mit
# see https://gitlab.com/pdbjapan/tools/cif-parsers/blob/master/license
import contextlib
import gzip
import json
import re


def partition_string(string, sep):
    return string.partition(sep)


class LoopMMCIF:
    def __init__(self, parser_obj):
        self.parser_obj = parser_obj
        self.length = 0
        self.ref_id = -1
        self.ref_list = []
        self.names_defined = False

    def add_name(self, name):
        cat_name = type(name) == str and partition_string(name, ".") or ["", "", ""]
        if cat_name[1]:
            if cat_name[0] not in self.parser_obj.current_target[-2]:
                self.parser_obj.current_target[-2][cat_name[0]] = {}
            if cat_name[2] not in self.parser_obj.current_target[-2][cat_name[0]]:
                self.parser_obj.current_target[-2][cat_name[0]][cat_name[2]] = []
            self.ref_list.append(
                self.parser_obj.current_target[-2][cat_name[0]][cat_name[2]]
            )
        else:
            if cat_name[0] not in self.parser_obj.current_target[-2]:
                self.parser_obj.current_target[-2][cat_name[0]] = []
            self.ref_list.append(self.parser_obj.current_target[-2][cat_name[0]])
        self.length = len(self.ref_list)

    def push_value(self, value):
        if not self.names_defined:
            self.names_defined = True
        target = self.next_target()
        if value == "stop_":
            return self.stop_push()
        target.append(value)

    def next_target(self):
        self.ref_id = (self.ref_id + 1) % self.length
        return self.ref_list[self.ref_id]

    def stop_push(self):
        self.ref_id = -1


def special_split(content):
    output = [["", False]]
    quote = False
    length = len(content)
    for c in range(length):
        is_ws = content[c] in [" ", "\t"]
        if content[c] in ["'", '"'] and (
            (
                c == 0
                or content[c - 1] == " "
                or content[c - 1] == "\t"
                or c == length - 1
                or content[c + 1] == " "
                or content[c + 1] == "\t"
            )
        ):
            quote = not quote
        elif not quote and is_ws and output[-1][0] != "":
            output.append(["", False])
        elif not quote and content[c] == "#":
            break
        elif not is_ws or quote:
            output[-1][0] += content[c]
            output[-1][1] = quote
    if output[-1][0] == "":
        output.pop()
    return output


class TargetSetter:
    def __init__(self, obj, key):
        self.obj = obj
        self.key = key

    def set_value(self, value):
        self.obj[self.key] = value


class CIFParser:
    def __init__(self):
        self.data = {}
        self.current_target = None
        self.loop_pointer = None

    def parse_string(self, contents):
        multi_line_mode = False
        buffer = []
        for line in contents.splitlines():
            z = line[:1]
            line = line.strip()
            if z == ";":
                if multi_line_mode:
                    self.set_data_value("\n".join(buffer))
                else:
                    buffer = []
                multi_line_mode = not multi_line_mode
                line = line[1:].strip()
            if multi_line_mode:
                buffer.append(line)
            else:
                self.process_content(special_split(line))

    def parse(self, fileobj):
        multi_line_mode = False
        buffer = []
        for line in fileobj.readlines():
            z = line[:1]
            line = line.strip()
            if z == ";":
                if multi_line_mode:
                    self.set_data_value("\n".join(buffer))
                else:
                    buffer = []
                multi_line_mode = not multi_line_mode
                line = line[1:].strip()
            if multi_line_mode:
                buffer.append(line)
            else:
                self.process_content(special_split(line))

    def process_content(self, content):
        for c, quoted in content:
            if c == "global_" and not quoted:
                self.loop_pointer = None
                self.select_global()
            elif c[:5] == "data_" and not quoted:
                self.loop_pointer = None
                self.select_data(c)
            elif c[:5] == "save_" and not quoted:
                self.loop_pointer = None
                if c[5:]:
                    self.select_frame(c)
                else:
                    self.end_frame()
            elif c == "loop_" and not quoted:
                self.loop_pointer = LoopMMCIF(self)
            elif c[:1] == "_" and not quoted:
                self.set_data_name(c[1:])
            else:
                self.set_data_value(c)

    def set_data_name(self, name):
        if self.loop_pointer is not None:
            if self.loop_pointer.names_defined:
                self.loop_pointer = None
            else:
                return self.loop_pointer.add_name(name)
        name = partition_string(name, ".")
        self.current_target.pop()
        if name[1]:
            if name[0] not in self.current_target[-1]:
                self.current_target[-1][name[0]] = {}
            self.current_target[-1][name[0]][name[2]] = ""
            self.current_target = self.current_target + [
                TargetSetter(self.current_target[-1][name[0]], name[2])
            ]
        else:
            self.current_target[-1][name[0]] = ""
            self.current_target = self.current_target + [
                TargetSetter(self.current_target[-1], name[0])
            ]

    def set_data_value(self, value):
        if self.loop_pointer is None:
            self.current_target[-1].set_value([value])

        else:
            self.loop_pointer.push_value(value)

    def select_global(self):
        self.current_target = [self.data, self.data, None]

    def select_data(self, name):
        if name not in self.data:
            self.data[name] = {}
        self.current_target = [self.data, self.data[name], None]

    def select_frame(self, name=""):
        if name not in self.current_target[1]:
            self.current_target[1][name] = {}
        self.current_target = self.current_target[:2] + [
            self.current_target[1][name],
            None,
        ]

    def end_data(self):
        self.current_target = self.current_target[:2]

    def end_frame(self):
        self.current_target = self.current_target[:3]


class __CIFFloat__(float):
    def __repr__(self):
        return "%.15g" % self


class __CIFInt__(int):
    def __repr__(self):
        return str(self)


def __cif_float_range__(inp):
    try:
        pos = inp.index("-", 1)
        return (__CIFFloat__(inp[:pos]), __CIFFloat__(inp[pos + 1 :]))
    except:
        return (__CIFFloat__(inp),)


def __cif_int_range__(inp):
    try:
        pos = inp.index("-", 1)
        return (__CIFInt__(inp[:pos]), __CIFInt__(inp[pos + 1 :]))
    except:
        return (__CIFInt__(inp),)


def __load_cif_dic__(dic_file, force=False):
    jsf_dic = f"{dic_file[:-4]}.json"
    jsf = f"{dic_file[:-4]}_summary.json"
    dic = {}
    try:
        if force:
            throw
        dic = json.loads(open(jsf).read())
    except:
        parser = CIFParser()
        parser.parse(open(dic_file))
        json.dump(parser.data, open(jsf_dic, "w"))
        for k, v in parser.data["data_mmcif_pdbx.dic"].items():
            if type(v) != dict or "item_type" not in v:
                continue
            name = partition_string(k[6:], ".")
            if name[0] not in dic:
                dic[name[0]] = {}
            dic[name[0]][name[2]] = v["item_type"]["code"][0].strip()
        json.dump(dic, open(jsf, "w"))

    typing = {}
    for k, v in dic.items():
        for k2, v2 in v.items():
            if v2 == "float":
                if k not in typing:
                    typing[k] = {}
                typing[k][k2] = __CIFFloat__
            elif v2 == "float-range":
                if k not in typing:
                    typing[k] = {}
                typing[k][k2] = __cif_float_range__
            elif v2 == "int":
                if k not in typing:
                    typing[k] = {}
                typing[k][k2] = __CIFInt__
            elif v2 == "int-range":
                if k not in typing:
                    typing[k] = {}
                typing[k][k2] = __cif_int_range__
    return typing


def __dump_cif__(jso):
    return __dump_part__(jso)


__CIF_STR_CHECK__ = re.compile(r"[\\s\(\)]")
__CIF_STR_NL_CHECK__ = re.compile(r"[\n]")


def __dump_str__(inp):
    if inp is None:
        return "?"
    if type(inp) is not str:
        return str(inp)
    if re.search(__CIF_STR_NL_CHECK__, inp) is not None:
        return "\n;%s\n;" % inp
    return "'%s'" % inp if re.search(__CIF_STR_CHECK__, inp) is not None else inp


def __pad_string__(inp, flength):
    return inp + (" " * (flength - len(inp)))


def __dump_cat__(k, v):
    output = "#\n"
    noi = len(v[v.keys()[0]])
    if noi == 1:
        pad = 0
        for k2 in v.keys():
            if len(k2) > pad:
                pad = len(k2)
        pad += 3
        for k2 in v.keys():
            output += "_%s.%s%s\n" % (
                k,
                __pad_string__(k2, pad),
                __dump_str__(v[k2][0]),
            )
    else:
        output += "loop_\n"
        pad = []
        for k2 in v.keys():
            output += "_%s.%s\n" % (k, k2)
            pad.append(0)
        tmp1 = []
        for i in range(noi):
            tmp2 = []
            tmp1.append(tmp2)
            tmp2.extend(__dump_str__(v[k2][i]) for k2 in v.keys())
        for j in range(len(tmp1[0])):
            pad = 0
            for item_ in tmp1:
                if item_[j][:2] != "\n;" and len(item_[j]) > pad:
                    pad = len(item_[j])
            pad += 1
            for item in tmp1:
                if item[0][:2] != "\n;":
                    item[j] = __pad_string__(item[j], pad)

        for i in range(noi):
            output += "".join(tmp1[i]) + "\n"

    return output.strip() + "\n"


def __dump_part__(jso):
    inner = True
    output = ""
    for k, v in jso.items():
        if isinstance(v, dict):
            if k[:5] != "data_" and k[:5] != "save_" and k[:7] != "global_":
                output += __dump_cat__(k, v)
            else:
                output += k + "\n"
                output += __dump_part__(v)
                inner = False
    return output + "#\n" if inner else output


def load_cif_data(data, do_clean=True, do_type=True):
    parser = CIFParser()
    if type(data) == str:
        parser.parse_string(data)
    else:
        parser.parse(data)  # fileobj

    if not do_clean:
        return parser.data

    for k, v in parser.data.items():
        for k2, v2 in v.items():
            for k3, v3 in v2.items():
                for i in range(len(v3)):
                    v2[k3][i] = v3[i] not in ["?", "."] and v3[i] or None

    if not do_type or not __MMCIF_TYPING__:
        return parser.data

    for _, data in parser.data.items():
        for k, v in __MMCIF_TYPING__.items():
            if k not in data:
                continue
            for k2, v2 in v.items():
                if k2 in data[k]:
                    for r in range(len(data[k][k2])):
                        with contextlib.suppress(KeyError):
                            data[k][k2][r] = v2(data[k][k2][r])
    return parser.data


def __load_cif__(cif_file, do_clean=True, do_type=True):
    parser = CIFParser()
    if cif_file[-3:].lower() == ".gz":
        parser.parse(gzip.open(cif_file))
    else:
        parser.parse(open(cif_file))

    if not do_clean:
        return parser.data

    for k, v in parser.data.items():
        for k2, v2 in v.items():
            for k3, v3 in v2.items():
                for i in range(len(v3)):
                    v2[k3][i] = v3[i] not in ["?", "."] and v3[i]  # or None

    if not do_type or not __MMCIF_TYPING__:
        return parser.data

    for _, data in parser.data.items():
        for k, v in __MMCIF_TYPING__.items():
            if k not in data:
                continue
            for k2, v2 in v.items():
                if k2 in data[k]:
                    for r in range(len(data[k][k2])):
                        with contextlib.suppress(KeyError):
                            data[k][k2][r] = v2(data[k][k2][r])
    return parser.data


__MMCIF_TYPING__ = None



"""Class for working with MMTF files."""
from __future__ import annotations

import os
import copy
import gzip
import warnings
from string import ascii_uppercase
from typing import Any, Dict, List, Union
from warnings import warn

import numpy as np
import pandas as pd
from looseversion import LooseVersion
from mmtf import MMTFDecoder, MMTFEncoder, fetch, parse, parse_gzip

from biopandas.constants import protein_letters_3to1_extended

from ..pdb.engines import amino3to1dict, pdb_df_columns, pdb_records

pd_version = LooseVersion(pd.__version__)


class PandasMmtf(object):
    def __init__(self):
        self._df: Dict[str, pd.DataFrame] = {}
        self.mmtf = ""
        self.header = ""
        self.code = ""
        self._get_dict = {}
        self.mmtf_path = ""

    @property
    def df(self):
        """Access dictionary of pandas DataFrames for MMTF record sections."""
        return self._df

    @df.setter
    def df(self, value: Any):
        """Assign a new value to the pandas DataFrame"""
        raise AttributeError(
            "Please use `PandasMmtf._df = ... ` instead\n"
            "of `PandasMmtf.df = ... ` if you are sure that\n"
            "you want to overwrite the `df` attribute."
        )
        # self._df = value

    def read_mmtf(self, filename: Union[str, os.PathLike]):
        filename = str(filename)
        if filename.endswith(".gz"):
            self.mmtf = parse_gzip(filename)
        else:
            self.mmtf = parse(filename)
        self.mmtf_path = filename
        df = self._mmtf_to_df(self.mmtf)
        self._df["ATOM"] = df.loc[df.record_name == "ATOM"]
        self._df["HETATM"] = df.loc[df.record_name == "HETATM"]
        return self

    def fetch_mmtf(self, pdb_code: str):
        self.code = pdb_code
        self.mmtf = fetch(pdb_code)
        df = self._mmtf_to_df(self.mmtf)
        self._df["ATOM"] = df.loc[df.record_name == "ATOM"]
        self._df["HETATM"] = df.loc[df.record_name == "HETATM"]
        return self

    @staticmethod
    def _mmtf_to_df(mmtf_obj: MMTFDecoder) -> pd.DataFrame:
        return mmtf_to_df(mmtf_obj)

    def impute_element(self, records=("ATOM", "HETATM"), inplace=False):
        """Impute element_symbol from atom_name section.

        Parameters
        ----------
        records : iterable, default: ('ATOM', 'HETATM')
            Coordinate sections for which the element symbols should be
            imputed.

        inplace : bool, (default: False
            Performs the operation in-place if True and returns a copy of the
            MMTF DataFrame otherwise.

        Returns
        ---------
        DataFrame

        """
        if inplace:
            t = self.df
        else:
            t = self.df.copy()
            for d in self.df:
                t[d] = self.df[d].copy()

        for sec in records:
            t[sec]["element_symbol"] = t[sec][["atom_name", "element_symbol"]].apply(
                lambda x: x[0][1] if len(x[1]) == 3 else x[0][0], axis=1
            )
        return t

    @staticmethod
    def rmsd(df1, df2, s=None, invert=False):
        """Compute the Root Mean Square Deviation between molecules.

        Parameters
        ----------
        df1 : pandas.DataFrame
            DataFrame with HETATM, ATOM, and/or ANISOU entries.

        df2 : pandas.DataFrame
            Second DataFrame for RMSD computation against df1. Must have the
            same number of entries as df1.

        s : {'main chain', 'hydrogen', 'c-alpha', 'heavy', 'carbon'} or None,
            default: None
            String to specify which entries to consider. If None, considers
            all atoms for comparison.

        invert : bool, default: False
            Inverts the string query if true. For example, the setting
            `s='hydrogen', invert=True` computes the RMSD based on all
            but hydrogen atoms.

        Returns
        ---------
        rmsd : float
            Root Mean Square Deviation between df1 and df2

        """
        if df1.shape[0] != df2.shape[0]:
            raise AttributeError("DataFrames have unequal lengths")
        get_dict = PandasMmtf._init_get_dict()
        if s:
            if s not in get_dict.keys():
                raise AttributeError(f"s must be in {get_dict.keys()} or None")
            df1 = get_dict[s](df1, invert=invert)
            df2 = get_dict[s](df2, invert=invert)

        total = (
            (df1["x_coord"].values - df2["x_coord"].values) ** 2
            + (df1["y_coord"].values - df2["y_coord"].values) ** 2
            + (df1["z_coord"].values - df2["z_coord"].values) ** 2
        )
        return round((total.sum() / df1.shape[0]) ** 0.5, 4)

    @staticmethod
    def _init_get_dict():
        """Initialize dictionary for filter operations."""
        return {
            "main chain": PandasMmtf._get_mainchain,
            "hydrogen": PandasMmtf._get_hydrogen,
            "c-alpha": PandasMmtf._get_calpha,
            "carbon": PandasMmtf._get_carbon,
            "heavy": PandasMmtf._get_heavy,
        }

    @staticmethod
    def _get_mainchain(df, invert):
        """Return only main chain atom entries from a DataFrame"""
        if invert:
            mc = df[
                (df["atom_name"] != "C")
                & (df["atom_name"] != "O")
                & (df["atom_name"] != "N")
                & (df["atom_name"] != "CA")
            ]
        else:
            mc = df[
                (df["atom_name"] == "C")
                | (df["atom_name"] == "O")
                | (df["atom_name"] == "N")
                | (df["atom_name"] == "CA")
            ]
        return mc

    @staticmethod
    def _get_hydrogen(df, invert):
        """Return only hydrogen atom entries from a DataFrame"""
        if invert:
            return df[(df["element_symbol"] != "H")]
        else:
            return df[(df["element_symbol"] == "H")]

    @staticmethod
    def _get_heavy(df, invert):
        """Return only heavy atom entries from a DataFrame"""
        if invert:
            return df[df["element_symbol"] == "H"]
        else:
            return df[df["element_symbol"] != "H"]

    @staticmethod
    def _get_calpha(df, invert):
        """Return c-alpha atom entries from a DataFrame"""
        if invert:
            return df[df["atom_name"] != "CA"]
        else:
            return df[df["atom_name"] == "CA"]

    @staticmethod
    def _get_carbon(df, invert):
        """Return carbon atom entries from a DataFrame"""
        if invert:
            return df[df["element_symbol"] != "C"]
        else:
            return df[df["element_symbol"] == "C"]

    def amino3to1(self, record="ATOM", residue_col="residue_name", fillna="?"):
        """Creates 1-letter amino acid codes from DataFrame

        Non-canonical amino-acids are converted as follows:
        ASH (protonated ASP) => D
        CYX (disulfide-bonded CYS) => C
        GLH (protonated GLU) => E
        HID/HIE/HIP (different protonation states of HIS) = H
        HYP (hydroxyproline) => P
        MSE (selenomethionine) => M

        Parameters
        ----------
        record : str, default: 'ATOM'
            Specfies the record DataFrame.
        residue_col : str,  default: 'residue_name'
            Column in `record` DataFrame to look for 3-letter amino acid
            codes for the conversion.
        fillna : str, default: '?'
            Placeholder string to use for unknown amino acids.

        Returns
        ---------
        pandas.DataFrame : Pandas DataFrame object consisting of two columns,
            `'chain_id'` and `'residue_name'`, where the former contains
            the chain ID of the amino acid and the latter
            contains the 1-letter amino acid code, respectively.

        """
        tmp = self.df[record]
        cmp = "placeholder"
        indices = []

        residue_number_insertion = tmp["residue_number"].astype(str) + tmp["insertion"]

        for num, ind in zip(residue_number_insertion, np.arange(tmp.shape[0])):
            if num != cmp:
                indices.append(ind)
            cmp = num

        transl = tmp.iloc[indices][residue_col].map(amino3to1dict).fillna(fillna)

        return pd.concat((tmp.iloc[indices]["chain_id"], transl), axis=1)

    def distance(self, xyz=(0.00, 0.00, 0.00), records=("ATOM", "HETATM")):
        """Computes Euclidean distance between atoms and a 3D point.

        Parameters
        ----------
        xyz : tuple, default: (0.00, 0.00, 0.00)
            X, Y, and Z coordinate of the reference center for the distance
            computation.
        records : iterable, default: ('ATOM', 'HETATM')
            Specify which record sections to consider. For example, to consider
            both protein and ligand atoms, set `records=('ATOM', 'HETATM')`.
            This setting is ignored if `df` is not set to None.
            For downward compatibility, a string argument is still supported
            but deprecated and will be removed in future versions.

        Returns
        ---------
        pandas.Series : Pandas Series object containing the Euclidean
            distance between the atoms in the record section and `xyz`.

        """

        if isinstance(records, str):
            warnings.warn(
                "Using a string as `records` argument is "
                "deprecated and will not be supported in future"
                " versions. Please use a tuple or"
                " other iterable instead",
                DeprecationWarning,
            )
            records = (records,)

        df = pd.concat(objs=[self.df[i] for i in records])

        return np.sqrt(
            np.sum(
                df[["x_coord", "y_coord", "z_coord"]].subtract(xyz, axis=1) ** 2, axis=1
            )
        )

    @staticmethod
    def distance_df(df, xyz=(0.00, 0.00, 0.00)):
        """Computes Euclidean distance between atoms and a 3D point.

        Parameters
        ----------
        df : DataFrame
            DataFrame containing entries in the `PandasMmtf.df['ATOM']`
            or `PandasMmtf.df['HETATM']` format for the
            the distance computation to the `xyz` reference coordinates.
        xyz : tuple, default: (0.00, 0.00, 0.00)
            X, Y, and Z coordinate of the reference center for the distance
            computation.

        Returns
        ---------
        pandas.Series : Pandas Series object containing the Euclidean
            distance between the atoms in the record section and `xyz`.

        """
        return np.sqrt(
            np.sum(
                df[["x_coord", "y_coord", "z_coord"]].subtract(xyz, axis=1) ** 2, axis=1
            )
        )

    def to_pdb(self, path, records=None, gz=False, append_newline=True):
        """Write record DataFrames to a PDB file or gzipped PDB file.

        Parameters
        ----------
        path : str
            A valid output path for the pdb file

        records : iterable, default: None
            A list of PDB record sections in
            {'ATOM', 'HETATM', 'ANISOU', 'OTHERS'} that are to be written.
            Writes all lines to PDB if `records=None`.

        gz : bool, default: False
            Writes a gzipped PDB file if True.

        append_newline : bool, default: True
            Appends a new line at the end of the PDB file if True

        """
        if gz:
            openf = gzip.open
            w_mode = "wt"
        else:
            openf = open
            w_mode = "w"
        if not records:
            records = self.df.keys()

        dfs = {r: self.df[r].copy() for r in records if not self.df[r].empty}

        for r in dfs:
            for col in pdb_records[r]:
                dfs[r][col["id"]] = dfs[r][col["id"]].apply(col["strf"])
                dfs[r]["OUT"] = pd.Series("", index=dfs[r].index)

            for c in dfs[r].columns:
                # fix issue where coordinates with four or more digits would
                # cause issues because the columns become too wide
                if c in {"x_coord", "y_coord", "z_coord"}:
                    for idx in range(dfs[r][c].values.shape[0]):
                        if len(dfs[r][c].values[idx]) > 8:
                            dfs[r][c].values[idx] = str(dfs[r][c].values[idx]).strip()
                if c in {"line_idx", "OUT"}:
                    pass
                elif r in {"ATOM", "HETATM"} and c not in pdb_df_columns:
                    warn(
                        "Column %s is not an expected column and"
                        " will be skipped." % c
                    )
                else:
                    dfs[r]["OUT"] = dfs[r]["OUT"] + dfs[r][c]

        if pd_version < LooseVersion("0.17.0"):
            warn(
                "You are using an old pandas version (< 0.17)"
                " that relies on the old sorting syntax."
                " Please consider updating your pandas"
                " installation to a more recent version.",
                DeprecationWarning,
            )
            dfs.sort(columns="line_idx", inplace=True)

        elif pd_version < LooseVersion("0.23.0"):
            df = pd.concat(dfs)

        else:
            df = pd.concat(dfs, sort=False)

        df.sort_values(by="line_idx", inplace=True)

        with openf(path, w_mode) as f:

            s = df["OUT"].tolist()
            for idx in range(len(s)):
                if len(s[idx]) < 80:
                    s[idx] = f"{s[idx]}{' ' * (80 - len(s[idx]))}"
            to_write = "\n".join(s)
            f.write(to_write)
            if append_newline:
                f.write("\n")

    def parse_sse(self):
        """Parse secondary structure elements"""
        raise NotImplementedError

    def to_mmtf(self, path, records = ("ATOM", "HETATM")):
        """Write record DataFrames to an MMTF file.

        Parameters
        ----------
        path : str
            A valid output path for the mmtf file

        records: tuple(str):
            A tuple of records to write. Defaults to ("ATOM". "HETATM")

        """
        df = pd.concat(objs=[self.df[i] for i in records])
        return write_mmtf(df, path)


    def get_model(self, model_index: int) -> PandasMmtf:
        """Returns a new PandasPDB object with the dataframes subset to the
        given model index.

        Parameters
        ----------
        model_index : int
            An integer representing the model index to subset to.

        Returns
        ---------
        pandas_pdb.PandasPdb : A new PandasPdb object containing the
            structure subsetted to the given model.
        """

        df = copy.deepcopy(self)

        if "ATOM" in df.df.keys():
            df.df["ATOM"] = df.df["ATOM"].loc[df.df["ATOM"]["model_id"] == model_index]
        if "HETATM" in df.df.keys():
            df.df["HETATM"] = df.df["HETATM"].loc[
                df.df["HETATM"]["model_id"] == model_index
            ]
        if "ANISOU" in df.df.keys():
            df.df["ANISOU"] = df.df["ANISOU"].loc[
                df.df["ANISOU"]["model_id"] == model_index
            ]
        return df

    def get_models(self, model_indices: List[int]) -> PandasMmtf:
        """Returns a new PandasMmtf object with the dataframes subset to the
        given model index.

        Parameters
        ----------
        model_indices : List[int]
            A list representing the model indexes to subset to.

        Returns
        ---------
        pandas_pdb.PandasMmtf : A new PandasPdb object
            containing the structure subsetted to the given model.
        """

        df = copy.deepcopy(self)

        if "ATOM" in df.df.keys():
            df.df["ATOM"] = df.df["ATOM"].loc[
                [x in model_indices for x in df.df["ATOM"]["model_id"].tolist()]
            ]
        if "HETATM" in df.df.keys():
            df.df["HETATM"] = df.df["HETATM"].loc[
                [x in model_indices for x in df.df["HETATM"]["model_id"].tolist()]
            ]
        if "ANISOU" in df.df.keys():
            df.df["ANISOU"] = df.df["ANISOU"].loc[
                [x in model_indices for x in df.df["ANISOU"]["model_id"].tolist()]
            ]
        return df



def fetch_mmtf(pdb_code: str) -> pd.DataFrame:
    """Returns a dataframe from a PDB code.

    :param pdb_code: 4-letter PDB accession code
    :type pdb_code: str
    :return: Dataframe of protein structure
    :rtype: pd.DataFrame
    """
    df = fetch(pdb_code)
    return mmtf_to_df(df)


def parse_mmtf(file_path: str) -> pd.DataFrame:
    """Parses an MMTF file from disk and returns a pandas DataFrame.

    :param file_path: Path to mmtf file.
    :type file_path: str
    :return: Dataframe of protein structure.
    :rtype: pd.DataFrame
    """
    df = parse_gzip(file_path) if file_path.endswith(".gz") else parse(file_path)
    return mmtf_to_df(df)


def mmtf_to_df(mmtf_obj: MMTFDecoder) -> pd.DataFrame:
    data: Dict[str, Union[List[str], List[int], List[float], np.ndarray]] = {
        "record_name": [],
        "residue_name": [],
        "atom_name": [],
        "element_symbol": [],
        "charge": [],
        "x_coord": mmtf_obj.x_coord_list,
        "y_coord": mmtf_obj.y_coord_list,
        "z_coord": mmtf_obj.z_coord_list,
        "alt_loc": mmtf_obj.alt_loc_list,
        "b_factor": mmtf_obj.b_factor_list,
        "insertion": [],
        "residue_number": [],
        "occupancy": mmtf_obj.occupancy_list,
        "chain_id": [],
        "atom_number": mmtf_obj.atom_id_list,
        "model_id": [],
    }

    chain_indices = mmtf_obj.groups_per_chain

    for i, _ in enumerate(chain_indices):
        if i == 0:
            continue
        else:
            chain_indices[i] = chain_indices[i] + chain_indices[i - 1]
    model_indices = mmtf_obj.chains_per_model
    model_indices = [sum(model_indices[:i+1]) for i in range(len(model_indices))]
    ch_idx = 0

    entity_types = {}
    for i in mmtf_obj.entity_list:
        for chain in i["chainIndexList"]:
            entity_types[chain] = i["type"]

    ch_idx_iter = iter(sorted(list(entity_types.keys())))
    ch_idx = next(ch_idx_iter)
    for idx, i in enumerate(mmtf_obj.group_type_list):
        res = mmtf_obj.group_list[i]
        #record = "HETATM" if res["chemCompType"] == "NON-POLYMER" else "ATOM"
        #record = (
        #    "ATOM"
        #    if res["chemCompType"] in ["L-PEPTIDE LINKING", "PEPTIDE LINKING"]
        #    else "HETATM"
        #)
        if idx == chain_indices[ch_idx]:
            #ch_idx += 1
            ch_idx = next(ch_idx_iter)
        record = "ATOM" if entity_types[ch_idx] == "polymer" else "HETATM"

        for _ in res["atomNameList"]:
            data["residue_name"].append(res["groupName"])
            data["residue_number"].append(mmtf_obj.group_id_list[idx])
            #data["chain_id"].append([mmtf_obj.chain_name_list[ch_idx]])
            data["chain_id"].append([mmtf_obj.chain_name_list[ch_idx]])
            data["model_id"].append(int(np.argwhere(np.array(model_indices)>ch_idx)[0]) + 1)
            data["record_name"].append(record)
            data["insertion"].append(mmtf_obj.ins_code_list[idx])
        data["atom_name"].append(res["atomNameList"])
        data["element_symbol"].append(res["elementList"])
        data["charge"].append(res["formalChargeList"])

    for k, v in data.items():
        if k in [
            "residue_name",
            "x_coord",
            "y_coord",
            "z_coord",
            "alt_loc",
            "b_factor",
            "residue_number",
            "occupancy",
            "record_name",
            "insertion",
            "atom_number",
            "model_id"
        ]:
            continue
        data[k] = [i for sublist in v for i in sublist]

    df = pd.DataFrame.from_dict(data).sort_values(by=["model_id", "atom_number"])
    df.alt_loc = df.alt_loc.str.replace("\x00", "")
    df.insertion = df.insertion.str.replace("\x00", "")
    return df

def _seq1(seq, charmap: Dict[str, str], undef_code="X"):
    # sourcery skip: dict-assign-update-to-union
    """Convert protein sequence from three-letter to one-letter code.
    The single required input argument 'seq' should be a protein sequence
    using three-letter codes, either as a Python string or as a Seq or
    MutableSeq object.
    This function returns the amino acid sequence as a string using the one
    letter amino acid codes. Output follows the IUPAC standard (including
    ambiguous characters "B" for "Asx", "J" for "Xle", "X" for "Xaa", "U" for
    "Sel", and "O" for "Pyl") plus "*" for a terminator given the "Ter" code.
    Any unknown character (including possible gap characters), is changed
    into '-' by default.
    e.g.
    >>> from Bio.SeqUtils import seq1
    >>> seq1("MetAlaIleValMetGlyArgTrpLysGlyAlaArgTer")
    'MAIVMGRWKGAR*'
    The input is case insensitive, e.g.
    >>> from Bio.SeqUtils import seq1
    >>> seq1("METalaIlEValMetGLYArgtRplysGlyAlaARGTer")
    'MAIVMGRWKGAR*'
    You can set a custom translation of the codon termination code using the
    dictionary "custom_map" argument (defaulting to {'Ter': '*'}), e.g.
    >>> seq1("MetAlaIleValMetGlyArgTrpLysGlyAla***", custom_map={"***": "*"})
    'MAIVMGRWKGA*'
    You can also set a custom translation for non-amino acid characters, such
    as '-', using the "undef_code" argument, e.g.
    >>> seq1("MetAlaIleValMetGlyArgTrpLysGlyAla------ArgTer", undef_code='?')
    'MAIVMGRWKGA??R*'
    If not given, "undef_code" defaults to "X", e.g.
    >>> seq1("MetAlaIleValMetGlyArgTrpLysGlyAla------ArgTer")
    'MAIVMGRWKGAXXR*'
    """
    if charmap is None:
        charmap = {"Ter": "*"}
    # reverse map of threecode
    # upper() on all keys to enable caps-insensitive input seq handling
    onecode = {k.upper(): v for k, v in charmap.items()}
    # add the given termination codon code and custom maps
    onecode.update((k.upper(), v) for k, v in charmap.items())
    seqlist = [seq[3 * i : 3 * (i + 1)] for i in range(len(seq) // 3)]
    return "".join(onecode.get(aa.upper(), undef_code) for aa in seqlist)




def write_mmtf(df: pd.DataFrame, file_path: str):
    """Writes a biopandas dataframe to an MMTF file.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe to write
    file_path : str, default: '?'
        Path to output file.
    """
    count_models, count_chains, count_groups, count_atoms = 0, 0, 0, 0
    # Check if the input is a valid BioPandas DataFrame
    if not isinstance(df, pd.DataFrame):
        raise TypeError("The input must be a BioPandas DataFrame.")

    # Initialize MMTF encoder
    encoder = MMTFEncoder()
    encoder.init_structure(
        total_num_bonds=0,
        total_num_atoms=0,
        total_num_groups=0,
        total_num_chains=0,
        total_num_models=0,
        structure_id="",
    )
    encoder.set_xtal_info(space_group="", unit_cell=None)

    encoder.set_header_info(
        r_free=None,
        r_work=None,
        resolution=None,
        title=None,
        deposition_date=None,
        release_date=None,
        experimental_methods=None,
    )

    node_ids = df.model_id.astype(str) + ":" + df.chain_id + ":" + df.residue_name + ":" + df.residue_number.astype(str) + ":" + df.insertion.astype(str)
    df["residue_id"] = node_ids
    # Tracks values to replace them at the end
    chains_per_model = []
    groups_per_chain = []

    # Iterate over each model
    for model_idx, model in enumerate(df.model_id.unique()):
        # Subset the DataFrame to the current model
        # Ang get the chains
        model_df = df.loc[df.model_id == model]
        chains = model_df.chain_id.unique()

        count_models += 1
        # Set the model info
        encoder.set_model_info(
            #model_id=model_idx, # According to mmtf-python this is meaningless
            model_id=model_idx, # According to mmtf-python this is meaningless
            chain_count=0 # Set to 0 here and changed later
            )
        # Iterate over chains in model
        for chain_id in chains:
            seqs = []
            seq = ""
            prev_res_type = ""
            prev_resname = ""
            first_chain = True

            # Subset the dataframe to the current chain
            chain_df = model_df[model_df.chain_id == chain_id]

            # Iterate over residues in chain
            residues = chain_df.residue_id.unique()
            for residue_id in residues:
                count_groups += 1
                # Subset the dataframe to the current residue
                residue_df = chain_df.loc[chain_df.residue_id == residue_id]
                # Get residue and entity types
                if all(residue_df.record_name == "ATOM"):
                    residue_type = "ATOM"
                    entity_type = "polymer"
                elif all(residue_df.residue_name == "HOH"):
                    residue_type = "HETATM"
                    entity_type = "water"
                else:
                    residue_type = "HETATM"
                    entity_type = "non-polymer"
                # Get the 3-letter residue name
                resname = residue_df.residue_name.iloc[0]

                # Check if the molecule changes within the chain
                # This will always increment for the first residue in a
                #  chain due to the starting values above
                # Checking for similar entities is non-trivial from the
                #  structure object so we treat each molecule as a separate
                #  entity
                if residue_type != prev_res_type or (
                        residue_type == "HETATM" and resname != prev_resname
                    ):
                    encoder.set_entity_info(
                        chain_indices=[count_chains],
                        sequence="", # Set to empty here and changed later
                        description="",
                        entity_type=entity_type,
                    )
                    encoder.set_chain_info(
                        chain_id=chain_id,
                        chain_name="\x00" if len(chain_id.strip()) == 0 else chain_id,
                        num_groups=0, # Set to 0 here and changed later
                    )
                    if count_chains > 0:
                        groups_per_chain.append(count_groups - sum(groups_per_chain) -1)
                    if not first_chain:
                        seqs.append(seq)
                    first_chain = False
                    count_chains += 1
                    seq=""

                if entity_type == "polymer":
                    seq += _seq1(residue_df.residue_name.unique()[0], charmap=protein_letters_3to1_extended)

                prev_res_type = residue_type
                prev_resname = resname

                group_type = "NON-POLYMER" if residue_type == "HETATM" else "L-PEPTIDE LINKING"
                encoder.set_group_info(
                    group_name=residue_df.residue_name.unique()[0],
                    group_number=int(residue_df.residue_number.unique()[0]),
                    insertion_code="\x00" if residue_df.insertion.unique()[0] == "" else residue_df.insertion.unique()[0],
                    group_type=group_type, # Hack to ensure we can re-parse.
                    atom_count=len(residue_df),
                    bond_count=0,
                    single_letter_code=_seq1(df.residue_name.unique()[0], charmap=protein_letters_3to1_extended),
                    sequence_index=len(seq) - 1 if entity_type == "polymer" else -1,
                    secondary_structure_type=-1
                )
                for row in residue_df.itertuples():
                    count_atoms += 1
                    encoder.set_atom_info(
                        atom_name=row.atom_name,
                        serial_number=row.atom_number,
                        alternative_location_id="\x00" if row.alt_loc == "" else row.alt_loc,
                        x=row.x_coord,
                        y=row.y_coord,
                        z=row.z_coord,
                        occupancy=row.occupancy,
                        temperature_factor=row.b_factor,
                        element=row.element_symbol,
                        charge=0 if row.charge == "" else int(row.charge),
                    )
            seqs.append(seq)
            start_ind = len(encoder.entity_list) - len(seqs)
            for i, seq in enumerate(seqs):
                encoder.entity_list[start_ind + i]["sequence"] = seq

        chains_per_model.append(count_chains - sum(chains_per_model))
    groups_per_chain.append(count_groups - sum(groups_per_chain))

    encoder.chains_per_model = chains_per_model
    encoder.groups_per_chain = groups_per_chain
    encoder.num_atoms = count_atoms
    encoder.num_groups = count_groups
    encoder.num_chains = count_chains
    encoder.num_models = count_models
    encoder.finalize_structure()
    encoder.write_file(file_path)



# BioPandas
# Author: Arian Jamasb <arian@jamasb.io> Sebastian Raschka <mail@sebastianraschka.com>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas

import os

from biopandas.mmcif import PandasMmcif
from nose.tools import raises

TESTDATA_1t48 = os.path.join(os.path.dirname(__file__), "data", "1t48.cif")
TESTDATA_1t49 = os.path.join(os.path.dirname(__file__), "data", "1t49.cif")
# TESTDATA_lig1 = os.path.join(os.path.dirname(__file__), "data", "lig_conf_1.pdb")
# TESTDATA_lig2 = os.path.join(os.path.dirname(__file__), "data", "lig_conf_2.pdb")

TESTDATA_rna = os.path.join(os.path.dirname(__file__), "data", "1ehz.cif")

p1t48 = PandasMmcif()
p1t48.read_mmcif(TESTDATA_1t48)
p1t49 = PandasMmcif()
p1t49.read_mmcif(TESTDATA_1t49)

# pl1 = PandasPdb()
# pl1.read_pdb(TESTDATA_lig1)
# pl2 = PandasPdb()
# pl2.read_pdb(TESTDATA_lig2)


def test_equal():
    r = PandasMmcif.rmsd(p1t48.df["ATOM"], p1t48.df["ATOM"], s=None)
    assert r == 0.000, r


@raises(AttributeError)
def test_wrong_arg():
    PandasMmcif.rmsd(p1t48.df["ATOM"].loc[1:, :], p1t48.df["ATOM"], s="bla")


@raises(AttributeError)
def test_incompatible():
    PandasMmcif.rmsd(p1t48.df["ATOM"].loc[1:, :], p1t48.df["ATOM"], s=None)


@raises(AttributeError)
def test_invalid_query():
    PandasMmcif.rmsd(p1t48.df["ATOM"].loc[1:, :], p1t48.df["ATOM"], s="bla")


def test_protein():
    r = PandasMmcif.rmsd(p1t48.df["ATOM"], p1t49.df["ATOM"], s="c-alpha", invert=False)
    assert r == 0.4923, r


def test_rna_and_nonmatching_indices():
    ehz = PandasMmcif().read_mmcif(TESTDATA_rna)
    at = ehz.df["ATOM"]
    a64 = at[at["auth_seq_id"] == 64]
    a66 = at[at["auth_seq_id"] == 66]
    r = PandasMmcif.rmsd(a64, a66)
    assert r == 10.2007, r


# def test_ligand():
#    r = PandasMmcif.rmsd(pl1.df["HETATM"], pl2.df["HETATM"], s="hydrogen", invert=True)
#    assert r == 1.9959, r


# def test_ligand_default():
#    r = PandasMmcif.rmsd(pl1.df["HETATM"], pl2.df["HETATM"], s=None)
#    assert r == 2.6444, r



# BioPandas
# Author: Sebastian Raschka <mail@sebastianraschka.com>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas


import os
from urllib.error import HTTPError
from urllib.request import urlopen
from pathlib import Path

import numpy as np
import pandas as pd
from biopandas.mmcif import PandasMmcif
from biopandas.pdb import PandasPdb
from biopandas.testutils import assert_raises
from nose.tools import raises
from pandas.testing import assert_frame_equal

TESTDATA_FILENAME = os.path.join(os.path.dirname(__file__), "data", "3eiy.cif")

# Not clear on how ANISOU records are handled in mmCIF files so skipping
# TESTDATA_FILENAME2 = os.path.join(
#    os.path.dirname(__file__), "data", "4eiy_anisouchunk.cif"
# )
TESTDATA_FILENAME2 = os.path.join(os.path.dirname(__file__), "data", "4eiy.cif")
TESTDATA_FILENAME_GZ = os.path.join(os.path.dirname(__file__), "data", "3eiy.cif.gz")

TESTDATA_FILENAME_AF2_V4 = os.path.join(
    os.path.dirname(__file__), "data", "AF-Q5VSL9-F1-model_v4.cif"
)
TESTDATA_FILENAME_AF2_V3 = os.path.join(
    os.path.dirname(__file__), "data", "AF-Q5VSL9-F1-model_v3.cif"
)

ATOM_DF_COLUMNS = [
    "B_iso_or_equiv",
    "Cartn_x",
    "Cartn_y",
    "Cartn_z",
    "auth_asym_id",
    "auth_atom_id",
    "auth_comp_id",
    "auth_seq_id",
    "group_PDB",
    "id",
    "label_alt_id",
    "label_asym_id",
    "label_atom_id",
    "label_comp_id",
    "label_entity_id",
    "label_seq_id",
    "occupancy",
    "pdbx_PDB_ins_code",
    "pdbx_PDB_model_num",
    "pdbx_formal_charge",
    "type_symbol",
]

ANISOU_DF_COLUMNS = [
    "id",
    "type_symbol",
    "pdbx_label_atom_id",
    "pdbx_label_alt_id",
    "pdbx_label_comp_id",
    "pdbx_label_asym_id",
    "pdbx_label_seq_id",
    "pdbx_PDB_ins_code",
    "U[1][1]",
    "U[2][2]",
    "U[3][3]",
    "U[1][2]",
    "U[1][3]",
    "U[2][3]",
    "pdbx_auth_seq_id",
    "pdbx_auth_comp_id",
    "pdbx_auth_asym_id",
    "pdbx_auth_atom_id",
]

with open(TESTDATA_FILENAME, "r") as f:
    three_eiy = f.read()

with open(TESTDATA_FILENAME2, "r") as f:
    four_eiy = f.read()

with open(TESTDATA_FILENAME_AF2_V4, "r") as f:
    af2_test_struct_v4 = f.read()

with open(TESTDATA_FILENAME_AF2_V3, "r") as f:
    af2_test_struct_v3 = f.read()



def test__read_pdb():
    """Test private _read_pdb"""
    ppdb = PandasMmcif()
    path, txt = ppdb._read_mmcif(TESTDATA_FILENAME)
    print(txt)
    assert txt == three_eiy


def test__read_pdb_raises():
    """Test private _read_pdb:
    Test if ValueError is raised for wrong file formats."""

    expect = (
        "Wrong file format; allowed file formats are "
        ".cif, .cif.gz, .mmcif, .mmcif.gz"
    )

    def run_code_1():
        PandasMmcif()._read_mmcif("protein.mol2")

    assert_raises(ValueError, expect, run_code_1)

    def run_code_2():
        PandasMmcif()._read_mmcif("protein.mol2.gz")

    assert_raises(ValueError, expect, run_code_2)


def test_fetch_pdb():
    """Test fetch_pdb"""

    try:
        ppdb = PandasMmcif()
        url, txt = ppdb._fetch_mmcif("3eiy")
    except (HTTPError, ConnectionResetError):
        url, txt = None, None
    if txt:  # skip if PDB down
        txt[:100] == three_eiy[:100]
        ppdb.fetch_mmcif("3eiy")
        assert ppdb.mmcif_text == txt
        assert ppdb.mmcif_path == "https://files.rcsb.org/download/3eiy.cif"


def test_fetch_af2():
    """Test fetch_af2"""
    # Test latest release
    try:
        ppdb = PandasMmcif()
        url, txt = ppdb._fetch_af2("Q5VSL9", af2_version=4)
    except (HTTPError, ConnectionResetError):
        url, txt = None, None
    if txt:  # skip if AF DB down
        txt[:100] == af2_test_struct_v4[:100]
        ppdb.fetch_mmcif(uniprot_id="Q5VSL9", source="alphafold2-v4")
        assert ppdb.mmcif_text == txt
        assert (
            ppdb.mmcif_path
            == "https://alphafold.ebi.ac.uk/files/AF-Q5VSL9-F1-model_v4.cif"
        )

    # Test legacy release
    try:
        ppdb = PandasMmcif()
        url, txt = ppdb._fetch_af2("Q5VSL9", af2_version=3)
    except (HTTPError, ConnectionResetError):
        url, txt = None, None
    if txt:  # skip if AF DB down
        txt[:100] == af2_test_struct_v3[:100]
        ppdb.fetch_mmcif(uniprot_id="Q5VSL9", source="alphafold2-v3")
        assert ppdb.mmcif_text == txt
        assert (
            ppdb.mmcif_path
            == "https://alphafold.ebi.ac.uk/files/AF-Q5VSL9-F1-model_v3.cif"
        )


def test__read_pdb_gz():
    """Test public _read_pdb with gzip files"""
    ppdb = PandasMmcif()
    path, txt = ppdb._read_mmcif(TESTDATA_FILENAME_GZ)
    assert txt == three_eiy


def test__construct_df():
    """Test pandas dataframe construction"""
    ppdb = PandasMmcif()
    dfs = ppdb._construct_df(three_eiy)
    # assert set(dfs.keys()) == {"OTHERS", "ATOM", "ANISOU", "HETATM"}
    # Currently don't parse OTHERS records as I'm not sure where they're located in mmCIF files
    assert set(dfs.keys()) == {"ATOM", "ANISOU", "HETATM"}
    assert set(dfs["ATOM"].columns) == set(ATOM_DF_COLUMNS)
    assert set(dfs["HETATM"].columns) == set(ATOM_DF_COLUMNS)
    assert set(dfs["ANISOU"].columns) == set(ANISOU_DF_COLUMNS)
    exp = pd.Series(
        [
            52.73,
            2.527,
            54.656,
            -1.667,
            "A",
            "N",
            "SER",
            2,
            "ATOM",
            1,
            None,
            "A",
            "N",
            "SER",
            1,
            "23",
            1.0,
            "None",
            1,
            None,
            "N",
        ],
        index=[
            "B_iso_or_equiv",
            "Cartn_x",
            "Cartn_y",
            "Cartn_z",
            "auth_asym_id",
            "auth_atom_id",
            "auth_comp_id",
            "auth_seq_id",
            "group_PDB",
            "id",
            "label_alt_id",
            "label_asym_id",
            "label_atom_id",
            "label_comp_id",
            "label_entity_id",
            "label_seq_id",
            "occupancy",
            "pdbx_PDB_ins_code",
            "pdbx_PDB_model_num",
            "pdbx_formal_charge",
            "type_symbol",
        ],
    )
    # There's some weird behaviour here I can't quite figure out.
    # Comparing the two series fails on exp.label_alt_id and exp.pdbx_formal_charge
    # However, if I compare these values directly, they are equal.
    # assert exp.equals(dfs["ATOM"].loc[0, :])
    for k, v in exp.items():
        assert dfs["ATOM"].loc[0, :][k] == v, k


def test_read_pdb():
    """Test public read_pdb"""
    ppdb = PandasMmcif()
    ppdb.read_mmcif(TESTDATA_FILENAME)
    assert ppdb.pdb_text == three_eiy
    assert ppdb.code == "3eiy", ppdb.code
    assert ppdb.mmcif_path == TESTDATA_FILENAME


def test_read_pdb_from_list():
    """Test public read_pdb_from_list"""

    for pdb_text, code in zip([three_eiy, four_eiy], ["3eiy", "4eiy"]):
        ppdb = PandasMmcif()
        ppdb.read_mmcif_from_list(pdb_text)
        assert ppdb.pdb_text == pdb_text
        assert ppdb.code == code
        assert ppdb.mmcif_path == ""


def test_read_pdb_with_pathlib():
    """Test public read_pdb with pathlib.Path object as input"""
    ppdb = PandasMmcif()
    ppdb.read_mmcif(Path(TESTDATA_FILENAME))
    assert ppdb.pdb_text == three_eiy
    assert ppdb.code == "3eiy", ppdb.code
    assert ppdb.mmcif_path == TESTDATA_FILENAME


# Again, not sure how ANISOU are handled, so skipping
# It seems like the whole PDB ATOM record is duplicated - whereas
# previously ANISOU records were only present for a subset of atoms
# def test_anisou_input_handling():
#    """Test public read_pdb"""
#    ppdb = PandasMmcif()
#    ppdb.read_pdb(TESTDATA_FILENAME2)
#    assert ppdb.pdb_text == four_eiy
#    assert ppdb.code == "4eiy", ppdb.code


@raises(AttributeError)
def test_get_exceptions():
    ppdb = PandasMmcif()
    ppdb.read_mmcif(TESTDATA_FILENAME)
    ppdb.get("main-chai")


def test_get_all():
    ppdb = PandasMmcif()
    ppdb.read_mmcif(TESTDATA_FILENAME)
    for i in ["c-alpha", "hydrogen", "main chain"]:
        ppdb.get(i)


def test_get_df():
    ppdb = PandasMmcif()
    ppdb.read_mmcif(TESTDATA_FILENAME)

    shape = ppdb.get("c-alpha").shape
    assert shape == (174, 21), shape

    shape = ppdb.get("hydrogen", invert=True, records=("ATOM",)).shape
    assert shape == (1330, 21), shape

    # deprecated use of string
    shape = ppdb.get("hydrogen", invert=True, records="ATOM").shape
    assert shape == (1330, 21), shape

    shape = ppdb.get("hydrogen").shape
    assert shape == (0, 21), shape

    shape = ppdb.get("main chain", records=("ATOM",)).shape
    assert shape == (696, 21), shape

    shape = ppdb.get("heavy", records=("ATOM",)).shape
    assert shape == (1330, 21), shape

    shape = ppdb.get("carbon", records=("ATOM",)).shape
    assert shape == (857, 21), shape


def test_mmcif_pdb_conversion():
    """Tests conversion from mmCIF df to PDB df"""
    # Multichain test
    pdb = PandasPdb().fetch_pdb("4hhb")
    mmcif = PandasMmcif().fetch_mmcif("4hhb")
    mmcif_pdb = mmcif.convert_to_pandas_pdb()

    assert_frame_equal(
        pdb.df["ATOM"].drop(columns=["line_idx"]),
        mmcif_pdb.df["ATOM"].drop(columns=["line_idx"]),
    )
    assert_frame_equal(
        pdb.df["HETATM"].drop(columns=["line_idx"]),
        mmcif_pdb.df["HETATM"].drop(columns=["line_idx"]).reset_index(drop=True),
    )

    # single chain test
    pdb = PandasPdb().fetch_pdb("3eiy")
    mmcif = PandasMmcif().fetch_mmcif("3eiy")
    mmcif_pdb = mmcif.convert_to_pandas_pdb()

    assert_frame_equal(
        pdb.df["ATOM"].drop(columns=["line_idx"]),
        mmcif_pdb.df["ATOM"].drop(columns=["line_idx"]),
    )
    assert_frame_equal(
        pdb.df["HETATM"].drop(columns=["line_idx"]),
        mmcif_pdb.df["HETATM"].drop(columns=["line_idx"]).reset_index(drop=True),
    )
